"""Tool for searching the web for answers."""
from typing import Any, List, Optional, Union

from pydantic import Field

from steamship import Block, File, PluginInstance, SteamshipError, Task
from steamship.agents.schema import AgentContext, Tool
from steamship.data import TagValueKey
from steamship.experimental.easy.tags import get_tag_value_key
from steamship.utils.kv_store import KeyValueStore
from steamship.utils.repl import ToolREPL


class SearchTool(Tool):
    """
    Tool which uses Steamship's managed SERP API client to search Google.
    """

    name: str = "SearchTool"
    human_description: str = "Searches the web."
    agent_description: str = "Used to search the web for new information."

    cache: bool = False
    cache_store: Optional[KeyValueStore] = Field(None, exclude=True)

    class Config:
        arbitrary_types_allowed = True

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        """Execute a search using the Steamship plugin."""
        search_tool = context.client.use_plugin("serpapi-wrapper")

        if self.cache:
            self.cache_store = KeyValueStore(
                client=context.client, store_identifier="search-tool-serpapi-wrapper"
            )

        output = []
        for block in tool_input:
            if block.is_text():
                result = self._do_search(block.text, search_tool)
                if isinstance(result, str):
                    output.append(Block(text=result))
                else:
                    output.append(Block(text=f"{result}"))
        return output

    def _do_search(self, query: str, search_tool: PluginInstance) -> str:
        try:
            if (
                self.cache
            ):  # Not self.cache_store is not None, because it's set to a Field by default
                value = self.cache_store.get(query)
                if value is not None:
                    return value.get(TagValueKey.STRING_VALUE, "")

            task = search_tool.tag(doc=query)
            task.wait()
            answer = self._first_tag_value(
                # TODO: TagKind.SEARCH_RESULT
                task.output.file,
                "search-result",
                TagValueKey.STRING_VALUE,
            )

            if (
                self.cache
            ):  # Not self.cache_store is not None, because it's set to a Field by default
                self.cache_store.set(key=query, value={TagValueKey.STRING_VALUE: answer})

            return answer
        except SteamshipError:
            return "No search result found"

    @staticmethod
    def _first_tag_value(file: File, tag_kind: str, value_key: str) -> Optional[Any]:
        """Return the value of the first block tag found in a file for the kind and value_key specified."""
        for block in file.blocks:
            val = get_tag_value_key(block.tags, value_key, kind=tag_kind)
            if val is not None:
                return val
        return None


if __name__ == "__main__":
    ToolREPL(SearchTool()).run()



import contextlib
import random
import string

from steamship import EmbeddingIndex, Steamship, Workspace


def random_name() -> str:
    """Returns a random name suitable for a handle that has low likelihood of colliding with another.

    Output format matches test_[a-z0-9]+, which should be a valid handle.
    """
    letters = string.digits + string.ascii_letters
    return f"test_{''.join(random.choice(letters) for _ in range(10))}".lower()  # noqa: S311


_TEST_EMBEDDER = "test-embedder-v1"


@contextlib.contextmanager
def random_index(steamship: Steamship, plugin_instance: str) -> EmbeddingIndex:
    index = steamship.use_plugin(
        "embedding-index",
        random_name(),
        config={"embedder": {"plugin_handle": plugin_instance, "fetch_if_exists": True}},
        fetch_if_exists=True,
    )
    yield index
    index.delete()  # or whatever you need to do at exit


@contextlib.contextmanager
def temporary_workspace(steamship: Steamship) -> Workspace:
    workspace = Workspace.create(client=steamship)
    yield workspace
    workspace.delete()  # or whatever you need to do at exit



from __future__ import annotations

from typing import List

from steamship.plugin.outputs.plugin_output import PluginOutput


class EmbeddedItemsPluginOutput(PluginOutput):
    embeddings: List[List[float]]



"""Tool for generating images."""
from steamship.agents.tools import ImageGeneratorTool
from steamship.utils.repl import ToolREPL


class GoogleImageSearchTool(ImageGeneratorTool):
    """Tool to generate images from text by using Google Image Search."""

    name: str = "GoogleImageSearchTool"
    human_description: str = "Fetches an image from Google Image Search."
    agent_description = (
        "Used to retrieve an image of a well known object, person, place, or idea. Only use if the user has asked directly for an image. "
        "Input: a plain text string that describes an object, person, place, or idea. "
        "Output: an image of that thing."
    )
    generator_plugin_handle: str = "google-image-search"
    generator_plugin_config: dict = {}


if __name__ == "__main__":
    ToolREPL(GoogleImageSearchTool()).run()



import logging
import uuid
from abc import ABC, abstractmethod

from steamship import SteamshipError
from steamship.data.block import Block, BlockUploadType
from steamship.data.workspace import SignedUrl
from steamship.invocable import InvocableResponse, post
from steamship.invocable.plugin_service import PluginRequest, PluginService, TrainablePluginService
from steamship.plugin.inputs.raw_block_and_tag_plugin_input import RawBlockAndTagPluginInput
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.raw_block_and_tag_plugin_output import RawBlockAndTagPluginOutput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import TrainingParameterPluginOutput
from steamship.plugin.trainable_model import TrainableModel

# Note!
# =====
#
# This is the PLUGIN IMPLEMENTOR's View of a Generator.
#
# If you are using the Steamship Client, you probably want steamship.client.operations.generator instead
# of this file.
#
from steamship.utils.signed_urls import upload_to_signed_url


class Generator(PluginService[RawBlockAndTagPluginInput, RawBlockAndTagPluginOutput], ABC):
    @abstractmethod
    def run(
        self, request: PluginRequest[RawBlockAndTagPluginInput]
    ) -> InvocableResponse[RawBlockAndTagPluginOutput]:
        raise NotImplementedError()

    @post("generate")
    def run_endpoint(self, **kwargs) -> InvocableResponse[RawBlockAndTagPluginOutput]:
        """Exposes the Tagger's `run` operation to the Steamship Engine via the expected HTTP path POST /tag"""
        input = PluginRequest[RawBlockAndTagPluginInput].parse_obj(kwargs)
        for block in input.data.blocks:
            block.client = self.client
        result = self.run(input)

        # Rewrite block output by changing any blocks with byte content to pass by URL
        if result.data is not None and result.data.blocks is not None:
            result_blocks = []
            for block in result.data.blocks:
                if block.upload_type == BlockUploadType.FILE:
                    result_blocks.append(self.upload_block_content_to_signed_url(block))
                else:
                    result_blocks.append(block)
            result.data.blocks = result_blocks
        return result

    def upload_block_content_to_signed_url(self, block: Block) -> Block:
        """Recreate the block (create request) as a URL request, rather than direct content, since we can't do a multipart
        file upload from here."""
        if block.upload_bytes is None:
            raise SteamshipError(
                "There was an error with the plugin. When returning upload type FILE, the content may not be None."
            )
        filepath = str(uuid.uuid4())
        signed_url = (
            self.client.get_workspace()
            .create_signed_url(
                SignedUrl.Request(
                    bucket=SignedUrl.Bucket.PLUGIN_DATA,
                    filepath=filepath,
                    operation=SignedUrl.Operation.WRITE,
                )
            )
            .signed_url
        )

        logging.info(f"Got signed url for uploading block content: {signed_url}")

        upload_to_signed_url(signed_url, block.upload_bytes)

        read_signed_url = (
            self.client.get_workspace()
            .create_signed_url(
                SignedUrl.Request(
                    bucket=SignedUrl.Bucket.PLUGIN_DATA,
                    filepath=filepath,
                    operation=SignedUrl.Operation.READ,
                )
            )
            .signed_url
        )

        return Block(
            url=read_signed_url,
            upload_type=BlockUploadType.URL,
            mime_type=block.mime_type,
            tags=block.tags,
            text=block.text,
        )


class TrainableGenerator(
    TrainablePluginService[RawBlockAndTagPluginInput, RawBlockAndTagPluginOutput], ABC
):
    @abstractmethod
    def run_with_model(
        self, request: PluginRequest[RawBlockAndTagPluginInput], model: TrainableModel
    ) -> InvocableResponse[RawBlockAndTagPluginOutput]:
        raise NotImplementedError()

    # noinspection PyUnusedLocal
    @post("generate")
    def run_endpoint(self, **kwargs) -> InvocableResponse[RawBlockAndTagPluginOutput]:
        """Exposes the Tagger's `run` operation to the Steamship Engine via the expected HTTP path POST /generate"""
        return self.run(PluginRequest[RawBlockAndTagPluginInput].parse_obj(kwargs))

    # noinspection PyUnusedLocal
    @post("getTrainingParameters")
    def get_training_parameters_endpoint(
        self, **kwargs
    ) -> InvocableResponse[TrainingParameterPluginOutput]:
        """Exposes the Service's `get_training_parameters` operation to the Steamship Engine via the expected HTTP path POST /getTrainingParameters"""
        return self.get_training_parameters(PluginRequest[TrainingParameterPluginInput](**kwargs))

    # noinspection PyUnusedLocal
    @post("train")
    def train_endpoint(self, **kwargs) -> InvocableResponse[TrainPluginOutput]:
        """Exposes the Service's `train` operation to the Steamship Engine via the expected HTTP path POST /train"""
        logging.info(f"Tagger:train_endpoint called. Calling train {kwargs}")
        arg = PluginRequest[TrainPluginInput].parse_obj(kwargs)
        model = self.model_cls()()
        model.receive_config(config=self.config)

        if arg.is_status_check:
            return self.train_status(arg, model)
        else:
            return self.train(arg, model)



from typing import Any, Dict, Optional

from steamship.utils.metadata import hash_dict


def safe_get(d: Dict, key: str, default: Any = None) -> Optional[Any]:
    """Safely a value from dictionairy using a specific key"""
    return d.get(key, default) or default


def format_uri(uri: Optional[str]) -> Optional[str]:
    if uri is not None and not uri.endswith("/"):
        uri += "/"
    return uri


def create_instance_handle(
    invocable_handle: str, version_handle: str, invocable_config: Optional[Dict[str, Any]]
) -> str:
    """Create an instance handle deterministically from package/plugin metadata and configuration."""
    return f"{invocable_handle}-{hash_dict({**invocable_config, 'version': version_handle})}"



import json
from abc import abstractmethod
from typing import Any, List, Optional, Union

import requests

from steamship import Block, File, MimeTypes, Task
from steamship.agents.schema import AgentContext, Tool


class GeneratorTool(Tool):
    """
    A base class for tools that wrap Steamship Generator plugins. Subclass this and implement
    the `accept_output_block` method.
    """

    generator_plugin_handle: str
    generator_plugin_instance_handle: Optional[str] = None
    generator_plugin_config: dict = {}
    merge_blocks: bool = False

    @abstractmethod
    def accept_output_block(self, block: Block) -> bool:
        raise NotImplementedError()

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        generator = context.client.use_plugin(
            plugin_handle=self.generator_plugin_handle,
            instance_handle=self.generator_plugin_instance_handle,
            config=self.generator_plugin_config,
        )

        tasks = []

        if self.merge_blocks:
            block = tool_input[0]
            for extra_block in tool_input[1:]:
                block.text = f"{block.text}\n\n{extra_block.text}"
            tool_input = [block]

        for block in tool_input:
            if isinstance(block, str):
                print(block)
            if not block.is_text():
                continue

            prompt = block.text
            task = generator.generate(text=prompt, append_output_to_file=True)
            tasks.append(task)

        # TODO / REMOVE Synchronous execution is a temporary simplification while we merge code.
        output_blocks = []
        # NB: In an async framework, we need to contend with the fact that we have some Monad-style trickery
        # to contend with. Namely that the below code has taken us from List[Block] as the universal type to
        # List[List[Block]] as the async result that we will have to process.
        #
        # It's unclear to me (ted) if this is something we leave as an exercise for implementors or build in
        # as universally handled.
        for task in tasks:
            task.wait()
            task_blocks = self.post_process(task, context)
            for block in task_blocks:
                output_blocks.append(block)
        # END TODO / REMOVE

        return output_blocks

    def post_process(self, task: Task, context: AgentContext) -> List[Block]:
        """In this case, the Generator returns a GeneratorResponse that has a .blocks method on it"""
        try:
            return task.output.blocks
        except BaseException:
            return json.loads(task.output).get("blocks")


class ImageGeneratorTool(GeneratorTool):
    """
    A base class for tools that wrap Steamship Image Generator plugins.
    """

    def accept_output_block(self, block: Block) -> bool:
        return block.is_image()


class AudioGeneratorTool(GeneratorTool):
    """
    A base class for tools that wrap Steamship Audio Generator plugins.
    """

    def accept_output_block(self, block: Block) -> bool:
        return block.is_audio()


class ScrapeAndBlockifyTool(Tool):
    """
    A base class for tools that wrap Steamship Blockifier plugin which transforms bytes to a set of blocks.
    """

    blockifier_plugin_handle: str
    blockifier_plugin_instance_handle: Optional[str] = None
    blockifier_plugin_config: dict = {}

    @abstractmethod
    def should_blockify(self, block: Block) -> bool:
        raise NotImplementedError()

    def get_mime_type(self):
        return None

    def _scrape(self, url: str, context: AgentContext) -> File:
        response = requests.get(url)
        file = File.create(context.client, content=response.content, mime_type=self.get_mime_type())
        return file

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        tasks = []

        blockifier = context.client.use_plugin(
            plugin_handle=self.blockifier_plugin_handle,
            instance_handle=self.blockifier_plugin_instance_handle,
            config=self.blockifier_plugin_config,
        )

        for input_block in tool_input:
            if not input_block.is_text():
                continue

            url = input_block.text
            file = self._scrape(url, context)
            task = file.blockify(blockifier.handle)
            tasks.append(task)

        # TODO / REMOVE Synchronous execution is a temporary simplification while we merge code.
        output_blocks = []
        # NB: In an async framework, we need to contend with the fact that we have some Monad-style trickery
        # to contend with. Namely that the below code has taken us from List[Block] as the universal type to
        # List[List[Block]] as the async result that we will have to process.
        #
        # It's unclear to me (ted) if this is something we leave as an exercise for implementors or build in
        # as universally handled.
        for task in tasks:
            task.wait()
            task_blocks = self.post_process(task, context)
            for block in task_blocks:
                output_blocks.append(block)
        # END TODO / REMOVE

        return output_blocks

    def post_process(self, task: Task, context: AgentContext) -> List[Block]:
        """In this case, the Blockifier returns a BlockAndTagResponse that has a .file.blocks method on it"""
        try:
            return task.output.file.blocks
        except BaseException:
            return json.loads(task.output).get("file", {}).get("blocks")


class ImageBlockifierTool(ScrapeAndBlockifyTool):
    """
    A base class for tools that wrap Steamship Image Blockifier plugins.
    """

    def get_mime_type(self):
        return MimeTypes.PNG

    def should_blockify(self, block: Block) -> bool:
        return block.is_image()


class AudioBlockifierTool(ScrapeAndBlockifyTool):
    """
    A base class for tools that wrap Steamship Audio Blockifier plugins.
    """

    def get_mime_type(self):
        return MimeTypes.MP3

    def should_blockify(self, block: Block) -> bool:
        return block.is_audio()



import base64
import json

from steamship.base.model import CamelModel
from steamship.plugin.outputs.raw_data_plugin_output import RawDataPluginOutput
from steamship.utils.binary_utils import flexi_create


def test_dump_string():
    data, mime, encoding = flexi_create(string="Hi")
    assert data == "Hi"


def test_dump_json():
    assert json.dumps(flexi_create(json="Hi")[0]) == '"Hi"'
    assert json.dumps(flexi_create(json=True)[0]) == "true"
    assert json.dumps(flexi_create(json=1.2)[0]) == "1.2"
    assert json.dumps(flexi_create(json=[1, 2, 3])[0]) == "[1, 2, 3]"
    assert json.dumps(flexi_create(json={"hi": "there"})[0]) == '{"hi": "there"}'

    class Person(CamelModel):
        name: str

    person = Person(name="Ted")
    assert json.dumps(flexi_create(json=person)[0]) == '{"name": "Ted"}'


def _base64_decode(base64_message: str) -> str:
    base64_bytes = base64_message.encode("ascii")
    message_bytes = base64.b64decode(base64_bytes)
    return message_bytes.decode("utf8")


def test_dump_raw_data():
    """This makes use of flexi_create. We want to ensure it is properly base64 encoding the JSON representation of
    objects given to it."""

    obj = RawDataPluginOutput(json={"hi": "there"})
    json_str = _base64_decode(obj.data)
    assert json_str == '{"hi": "there"}'

    class Person(CamelModel):
        name: str

    person = Person(name="Ted")
    obj2 = RawDataPluginOutput(json=person)
    json_str2 = _base64_decode(obj2.data)
    assert json_str2 == '{"name": "Ted"}'



import os
from pathlib import Path

from assets.plugins.taggers.plugin_trainable_tagger import (
    TRAINING_PARAMETERS,
    TestTrainableTaggerModel,
)
from steamship_tests.utils.fixtures import get_steamship_client

from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.outputs.model_checkpoint import ModelCheckpoint
from steamship.plugin.request import PluginRequest


def _test_folders_equal(p1: Path, p2: Path):
    assert p1 != p2

    p1_files = os.listdir(p1)
    p2_files = os.listdir(p2)

    assert p1_files == p2_files

    for file in p1_files:
        with open(p1 / file, "r") as f1:
            with open(p2 / file, "r") as f2:
                assert f1.read() == f2.read()


def test_model_checkpoint_save_load():
    """A ModelCheckpoint captures the entire state of a model at any given time.

    On disk, it is a folder.
    - Models are expected to take this folder as their initialization input
    - Models are expected to persist whatever they need to persist to this folder during training.

    On Steamship, it is a zip archive stored in the associated PluginInstance's "Workspace Bucket"
    - Each ModelCheckpoint uploaded has a handle (like V1)
    - Every ModelCheckpoint uploaded can also become the new default
    """
    client = get_steamship_client()

    checkpoint_1 = ModelCheckpoint(client=client, handle="epoch1", plugin_instance_id="0000")
    with open(checkpoint_1.folder_path_on_disk() / "params.json", "w") as f:
        f.write("HI THERE")
    checkpoint_1.upload_model_bundle()

    # Now let's download the checkpoint labeled "epoch1" and test that it is equal
    checkpoint_downloaded = ModelCheckpoint(
        client=client, handle="epoch1", plugin_instance_id="0000"
    )
    checkpoint_downloaded.download_model_bundle()
    _test_folders_equal(
        checkpoint_1.folder_path_on_disk(), checkpoint_downloaded.folder_path_on_disk()
    )

    # We should also be able to download "default" checkpoint by not providing a handle
    checkpoint_default_1 = ModelCheckpoint(client=client, plugin_instance_id="0000")
    checkpoint_default_1.download_model_bundle()
    _test_folders_equal(
        checkpoint_1.folder_path_on_disk(), checkpoint_default_1.folder_path_on_disk()
    )

    # Let's create a new checkpoint with our trainer... epoch2
    checkpoint_2 = ModelCheckpoint(client=client, handle="epoch2", plugin_instance_id="0000")
    with open(checkpoint_2.folder_path_on_disk() / "params.json", "w") as f:
        f.write("UPDATED PARAMS")
    checkpoint_2.upload_model_bundle()

    # If we download the new DEFAULT checkpoint, we will now receive the epoch2 files...
    checkpoint_default_2 = ModelCheckpoint(client=client, plugin_instance_id="0000")
    checkpoint_default_2.download_model_bundle()
    _test_folders_equal(
        checkpoint_2.folder_path_on_disk(), checkpoint_default_2.folder_path_on_disk()
    )


def test_model_can_save_to_and_load_from_checkpoint():
    """Elaboration of model_checkpoint_save_load in which include an example Model in the loop."""

    client = get_steamship_client()
    plugin_instance_id = "000"

    # TRAIN PHASE #1
    # ====================================

    # Create a new, empty checkpoint

    # Create a new model. Train it.
    model = TestTrainableTaggerModel()
    model.train(
        PluginRequest(
            data=TrainPluginInput(plugin_instance="foo", training_params=TRAINING_PARAMETERS)
        )
    )

    # Create a checkpoint. Save the model to it. Upload it.
    model.save_remote(client=client, plugin_instance_id=plugin_instance_id, checkpoint_handle="V1")

    # INFERENCE PHASE #1
    # ====================================

    # Now we'll create two new copies of the model
    default_model = TestTrainableTaggerModel.load_remote(
        client=client, plugin_instance_id=plugin_instance_id
    )
    v1_model = TestTrainableTaggerModel.load_remote(
        client=client, plugin_instance_id=plugin_instance_id, checkpoint_handle="V1"
    )

    # Verify none of these models share the same state folder on disk.
    assert default_model.path != model.path
    assert v1_model.path != model.path
    assert v1_model.path != default_model.path

    # Verify all of these models share the same parameters.
    # V1 and Default were both loaded from the ModelCheckpoint!
    assert model.keyword_list == default_model.keyword_list
    assert v1_model.keyword_list == default_model.keyword_list
    assert v1_model.keyword_list == default_model.keyword_list

    # Let's do some more manual testing of that checkpoint..
    assert os.path.exists(model.path / TestTrainableTaggerModel.KEYWORD_LIST_FILE)
    assert os.path.exists(default_model.path / TestTrainableTaggerModel.KEYWORD_LIST_FILE)
    assert os.path.exists(v1_model.path / TestTrainableTaggerModel.KEYWORD_LIST_FILE)

    with open(model.path / TestTrainableTaggerModel.KEYWORD_LIST_FILE, "r") as f1:
        with open(default_model.path / TestTrainableTaggerModel.KEYWORD_LIST_FILE, "r") as f2:
            with open(v1_model.path / TestTrainableTaggerModel.KEYWORD_LIST_FILE, "r") as f3:
                original_model_params = f1.read()
                default_model_params = f2.read()
                v1_model_params = f3.read()
                assert original_model_params == default_model_params
                assert original_model_params == v1_model_params



from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union

from pydantic import Field
from pydantic.main import BaseModel

from steamship import SteamshipError
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import Request
from steamship.base.response import Response
from steamship.data.tags.tag_constants import GenerationTag, TagKind, TagValueKey


class TagQueryRequest(Request):
    tag_filter_query: str


class Tag(CamelModel):
    # Steamship client.
    client: Client = Field(None, exclude=True)

    # ID of the tag in the database.
    id: str = None

    # ID of the file associated with the tag.
    file_id: str = None

    # ID of the block associated with the tag. If not None, `start_idx` and `end_idx` should be set.
    block_id: Optional[str] = None

    # The kind of tag. See the ``TagKind`` enum class for suggestions.
    kind: Union[TagKind, str] = None  # E.g. ner

    # The name of tag. See the ``DocTag``, ``TokenTag``, etc enum classes for suggestions.
    name: Optional[str] = None  # E.g. person

    # The value payload of the tag. Always a JSON-style object.
    value: Optional[Dict[Union[TagValueKey, str], Any]] = None

    # Character index in associated block of the start of the span of text this tag comments upon. Start-inclusive.
    start_idx: Optional[int] = None

    # Character index in associated block of the end of the span of text this tag comments upon. End-exclusive.
    end_idx: Optional[int] = None

    # The text covered by the tag.
    # Note:
    #   The text will not always be materialized into the tag object
    #   itself; you may have to fetch it with file.text[tag.start_idx:tag.end_idx]
    # Note:
    #   Changing this field will not result in changes to Steamship's database.
    #   TODO(ted): Consider refactoring as a read-only property.
    #
    text: Optional[str] = None

    class DeleteRequest(Request):
        id: str = None
        file_id: str = None
        block_id: str = None

    class ListRequest(Request):
        file_id: str = None
        block_id: str = None

    class ListResponse(Response):
        tags: List[Tag] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["tag"] if "tag" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        file_id: str = None,
        block_id: str = None,
        kind: str = None,
        name: str = None,
        start_idx: int = None,
        end_idx: int = None,
        value: Dict[str, Any] = None,
    ) -> Tag:
        req = Tag(
            file_id=file_id,
            block_id=block_id,
            kind=kind,
            name=name,
            start_idx=start_idx,
            end_idx=end_idx,
            value=value,
        )
        return client.post("tag/create", req, expect=Tag)

    def delete(self) -> Tag:
        return self.client.post(
            "tag/delete",
            Tag.DeleteRequest(id=self.id, file_id=self.file_id, block_id=self.block_id),
            expect=Tag,
        )

    def index(self, plugin_instance: Any = None):
        """Index this tag."""
        return plugin_instance.insert(self)

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
    ) -> TagQueryResponse:
        req = TagQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "tag/query",
            payload=req,
            expect=TagQueryResponse,
        )
        return res


class TimestampTag(Tag):
    def __init__(
        self,
        start_time_s: float,
        end_time_s: float,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
        value: Optional[Dict[str, Any]] = None,
    ):
        value = value or {}
        super().__init__(
            kind=TagKind.TIMESTAMP,
            start_idx=start_idx,
            end_idx=end_idx,
            value={
                **value,
                TagValueKey.START_TIME_S: start_time_s,
                TagValueKey.END_TIME_S: end_time_s,
            },
        )


class TokenizationTag(Tag):
    class Type(str, Enum):
        PARAGRAPH = "paragraph"
        SENTENCE = "sentence"
        WORD = "word"
        CHARACTER = "character"

    def __init__(
        self,
        type=Type,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
        value: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            kind=TagKind.TOKENIZATION, name=type, start_idx=start_idx, end_idx=end_idx, value=value
        )


class SummaryTag(Tag):
    def __init__(
        self,
        summary: str,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
        value: Optional[Dict[str, Any]] = None,
    ):
        value = value or {}
        super().__init__(
            kind=TagKind.SUMMARY,
            start_idx=start_idx,
            end_idx=end_idx,
            value={**value, TagValueKey.VALUE: summary},
        )


class TopicTag(Tag):
    def __init__(
        self,
        topic: str,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
        value: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            kind=TagKind.TOPIC, name=topic, start_idx=start_idx, end_idx=end_idx, value=value
        )


class EmotionTag(Tag):
    class Emotion(str, Enum):
        POSITIVE = "positive"
        NEUTRAL = "neutral"
        NEGATIVE = "negative"
        HAPPINESS = "happiness"
        SADNESS = "sadness"
        JOY = "joy"
        LOVE = "love"
        ANGER = "anger"
        FEAR = "fear"
        SURPRISE = "surprise"
        HUMOR = "humor"
        CONCERN = "concern"
        SERIOUSNESS = "seriousness"
        SCORE = "score"

    def __init__(
        self,
        emotion: Emotion,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
        value: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            kind=TagKind.EMOTION, name=emotion, start_idx=start_idx, end_idx=end_idx, value=value
        )


class IntentTag(Tag):
    class Intent(str, Enum):
        SALUTATION = "salutation"
        PRAISE = "praise"
        COMPLAINT = "complaint"
        QUESTION = "question"
        REQUEST = "request"
        EXPLANATION = "explanation"
        SCHEDULING_REQUEST = "scheduling-request"
        ARE_YOU_THERE = "are-you-there"
        REVISITING_TOPIC = "revisiting-topic"

    def __init__(
        self,
        intent: Intent,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
        value: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            kind=TagKind.INTENT, name=intent, start_idx=start_idx, end_idx=end_idx, value=value
        )


class EntityTag(Tag):
    class EntityType(str, Enum):
        PERSON = "person"
        ORGANIZATION = "organization"
        PRODUCT = "product"
        LOCATION = "location"
        DATE = "date"
        TIME = "time"
        MONEY = "money"
        PERCENT = "percent"
        FACILITY = "facility"
        GEO_POLITICAL_ENTITY = "geo-political-entity"

    def __init__(
        self,
        entity_name: str,
        entity_type: EntityType,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
        value: Optional[Dict[str, Any]] = None,
    ):
        if TagValueKey.ENTITY_NAME in value:
            raise SteamshipError(
                f"The value of your EntityTag cannot contain the key {TagValueKey.ENTITY_NAME}."
            )
        super().__init__(
            kind=TagKind.ENTITY,
            name=entity_type,
            start_idx=start_idx,
            end_idx=end_idx,
            value={**value, TagValueKey.ENTITY_NAME: entity_name},
        )


class SentimentTag(Tag):
    class Sentiment(str, Enum):
        POSITIVE = "positive"
        NEUTRAL = "neutral"
        NEGATIVE = "negative"
        SCORE = "score"

    def __init__(
        self,
        sentiment: Sentiment,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
        value: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            kind=TagKind.ENTITY, name=sentiment, start_idx=start_idx, end_idx=end_idx, value=value
        )


class EmbeddingTag(Tag):
    def __init__(
        self,
        embedding: List[Union[float, int]] = None,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
        value: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            kind=TagKind.EMBEDDING,
            name=EmbeddingTag,
            start_idx=start_idx,
            end_idx=end_idx,
            value={**value, TagValueKey.VECTOR_VALUE: embedding},
        )


class PromptCompletionTag(Tag):
    def __init__(
        self,
        text: str = None,
        start_idx: Optional[int] = None,
        end_idx: Optional[int] = None,
        value: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(
            kind=TagKind.GENERATION,
            name=GenerationTag.PROMPT_COMPLETION,
            start_idx=start_idx,
            end_idx=end_idx,
            value={**value, TagValueKey.STRING_VALUE: text},
        )


class TagQueryResponse(Response):
    tags: List[Tag]


Tag.ListResponse.update_forward_refs()
TimestampTag.update_forward_refs()
TopicTag.update_forward_refs()
SummaryTag.update_forward_refs()
TokenizationTag.update_forward_refs()
SentimentTag.update_forward_refs()
EntityTag.update_forward_refs()
IntentTag.update_forward_refs()
EmotionTag.update_forward_refs()



import json
import logging
import os
from json import JSONDecodeError
from os import path
from pathlib import Path
from typing import Any, Callable, Dict, Optional

import click
from pydantic import ValidationError

from steamship import Steamship, SteamshipError
from steamship.cli.ship_spinner import ship_spinner
from steamship.data.manifest import DeployableType, Manifest
from steamship.utils.metadata import hash_dict
from steamship.utils.utils import create_instance_handle


class LoggingDisabled:
    """Context manager that turns off logging within context."""

    def __enter__(self):
        logging.disable(logging.CRITICAL)

    def __exit__(self, exit_type, exit_value, exit_traceback):
        logging.disable(logging.NOTSET)


@click.command(name="use")
@click.option(
    "--workspace",
    "-w",
    required=False,
    type=str,
    help="Workspace handle. The new instance will be created in this workspace. If not specified, "
    "the default workspace will be used.",
)
@click.option(
    "--instance",
    "-i",
    type=str,
    required=False,
    help="Instance handle. Used to name the instance (if it needs to be created). If not specified, "
    "a generated instance name will be used.",
)
@click.option(
    "--config",
    "-c",
    type=str,
    required=False,
    help="Instance configuration. May be inline JSON or a path to a file. If not specified, "
    "an empty configuration dictionary will be passed to the instance.",
)
def create_instance(
    workspace: Optional[str] = None, instance: Optional[str] = None, config: Optional[str] = None
):
    """Create an instance of your package/plugin for use.

    Must be run from a directory containing a Steamship manifest.
    """

    with LoggingDisabled():  # keep CLI output as concise as possible
        return _create_instance(workspace, instance, config)


def _create_instance(  # noqa: C901
    workspace: Optional[str] = None,
    instance_handle: Optional[str] = None,
    config: Optional[str] = None,
):
    manifest = None
    if path.exists("steamship.json"):
        try:
            manifest = Manifest.load_manifest()
        except ValidationError as e:
            click.secho("")
            click.secho("This package had an invalid steamship.json file.", fg="red")
            click.secho("")
            click.secho(f"{e}", fg="red")
            click.secho("")
            click.get_current_context().abort()
    else:
        click.secho("No manifest found for instance creation.", fg="red")
        click.secho("Please try again from a directory with a package or plugin manifest.")
        click.get_current_context().abort()

    if manifest is None:
        click.secho("Steamship manifest failed to load.", fg="red")
        click.get_current_context().abort()

    invocable_config, is_file = config_str_to_dict(config)
    new_param_values = False
    for param, param_config in manifest.configTemplate.items():
        if param not in invocable_config and param_config.default is None:
            if param.upper() in os.environ:
                invocable_config[param] = os.environ[param.upper()]
            else:
                invocable_config[param] = click.prompt(
                    f"Value for {param} ({param_config.description})"
                )
            new_param_values = True

    if is_file and new_param_values:
        if click.confirm(
            f"Do you want to store this config in your config file ({config})?", default=True
        ):
            json.dump(invocable_config, Path(config).open("w"))
            click.secho(
                f"Successfully wrote configuration {json.dumps(config)} to {config}.", fg="green"
            )

    if workspace is None or len(workspace) == 0:
        ws_hash = hash_dict(
            {
                **invocable_config,
                "version": manifest.version,
                "invocable": manifest.handle,
            }
        )
        workspace = f"space-{ws_hash}"

    client = None
    try:
        client = Steamship(workspace=workspace)
    except SteamshipError as e:
        click.secho(e.message, fg="red")
        click.get_current_context().abort()

    create_instance_fn = client.use
    if manifest.type == DeployableType.PLUGIN:
        create_instance_fn = client.use_plugin

    if instance_handle is None:
        instance_handle = create_instance_handle(
            invocable_handle=manifest.handle,
            version_handle=manifest.version,
            invocable_config=invocable_config,
        )

    try:
        _call_create_instance_fn(
            client, manifest, instance_handle, invocable_config, create_instance_fn
        )
    except SteamshipError as e:
        if "Configuration for this PackageInstance is invalid." in e.message:
            click.secho(
                "\nThe configuration for this Package instance was invalid. "
                "This usually means it contains some required configuration fields.",
                fg="red",
            )
            click.secho("Create one interactively on the web at:\n", fg="red")
            click.secho(f"   https://steamship.com/packages/{manifest.handle}/_create\n")
        else:
            click.secho(f"\nFailed to create instance: {e.message}", fg="red")
        click.get_current_context().abort()


def _call_create_instance_fn(
    client: Steamship,
    manifest: Manifest,
    instance: str,
    config: Optional[Dict[str, Any]],
    create_instance_fn: Callable,
):
    """"""
    workspace_handle = client.get_workspace().handle

    click.echo("Using values:\n- Workspace: ", nl=False)
    click.secho(f"{workspace_handle}", fg="green")
    click.echo(f"- {manifest.type.title()}: ", nl=False)
    click.secho(f"{manifest.handle}", fg="green")
    click.echo("- Version: ", nl=False)
    click.secho(f"{manifest.version}", fg="green")
    click.echo("- Instance Handle: ", nl=False)
    click.secho(f"{instance}", fg="green")
    click.echo("- Configuration: ", nl=False)
    click.secho(f"{json.dumps(config)}", fg="green")

    click.echo("Creating new (or fetching existing) instance: ", nl=False)
    with ship_spinner():
        instance = create_instance_fn(
            manifest.handle,
            instance_handle=instance,
            version=manifest.version,
            config=config,
            wait_for_init=True,
        )
        if instance:
            click.secho("\nSuccess!", fg="green")
            click.secho("")
            if manifest.type == DeployableType.PACKAGE:
                click.echo(
                    f"Web URL: https://steamship.com/workspaces/{workspace_handle}/packages/{instance.handle}"
                )
                click.echo(f"API URL: {instance.invocation_url}")
                click.secho("")
            return
        else:
            raise SteamshipError("instance creation unexpectedly returned empty instance.")


def cannot_parse_config(config: str, exception: Optional[Exception] = None):
    exception_str = f": {exception}" if exception else ""
    click.secho(
        f"Could not parse configuration {config} as a file or JSON string{exception_str}",
        fg="red",
    )
    click.get_current_context().abort()


def config_str_to_dict(config: Optional[str] = None) -> (Dict[str, Any], bool):
    """Convert config string into dict.

    The input string can be a JSON-string or the name of a file."""

    if not config:
        return {}, False
    try:
        json_config = json.loads(config)
        if not isinstance(json_config, dict):
            cannot_parse_config(config)
        return json_config, False
    except JSONDecodeError:
        try:
            config_path = Path(config)
            if not config_path.exists():
                if click.confirm(
                    f"Configuration file {config_path} does not exist. Do you want to create it",
                    default=True,
                ):
                    config_path.parent.mkdir(parents=True, exist_ok=True)
                    json.dump({}, config_path.open("w"))
            file_json = json.load(config_path.open())
            if not isinstance(file_json, dict):
                cannot_parse_config(config)
            return file_json, True

        except Exception as e:
            cannot_parse_config(config, e)



from abc import abstractmethod
from typing import Any, List, Union

from pydantic.main import BaseModel

from steamship import Block, Task


class AgentContext(BaseModel):
    """Placeholder to avoid circular dependency."""

    # TODO(doug): refactor so this is not necessary
    pass


class Tool(BaseModel):
    """Tools provide functionality that may be used by `AgentServices`, as directed by `Agents`, to
    achieve a goal.

    Tools may be used to wrap Steamship packages and plugins, as well as third-party backend services,
    and even locally-contained bits of Python code.
    """

    name: str
    """The short name for the tool.
    This will be used by Agents to refer to this tool during action selection."""

    agent_description: str
    """Description for use in an agent in order to enable Action selection.
    It should include a short summary of what the Tool does, what the inputs to the Tool should be,
    and what the outputs of the tool are."""

    human_description: str
    """Human-friendly description.
    Used for logging, tool indices, etc."""

    @abstractmethod
    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        """Run the tool given the provided input and context.

        At the moment, only synchronous Tools (those that return List[Block]) are supported.

        Support for asynchronous Tools (those that return Task[Any]) will be added shortly.
        """
        raise NotImplementedError()

    def post_process(self, async_task: Task, context: AgentContext) -> List[Block]:
        """Transforms Task output into a List[Block]."""
        pass



"""Experimental Steamship Packages

This module contains experimental Steamship code. It's intended to serve as a place to store fast-moving
experiments that aren't yet ready for the core Steamship library.

In general, try to avoid introducing new dependencies to requirements.txt, but otherwise use this as a place
to try out new metaphors and helper classes!
"""

from .easy import blockify, scrape

__all__ = ["scrape", "blockify"]



from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.file import upload_file
from steamship_tests.utils.fixtures import get_steamship_client


def test_e2e_csv_blockifier_plugin():
    client = get_steamship_client()
    csv_blockifier_plugin_path = PLUGINS_PATH / "blockifiers" / "csv_blockifier.py"

    version_config_template = {
        "text_column": {"type": "string"},
        "tag_columns": {"type": "string"},
        "tag_kind": {"type": "string"},
    }  # TODO (enias): Derive this from Config
    instance_config = {  # Has to match up
        "text_column": "Message",
        "tag_columns": "Category",
        "tag_kind": "Intent",
    }
    with deploy_plugin(
        client,
        csv_blockifier_plugin_path,
        "blockifier",
        version_config_template=version_config_template,
        instance_config=instance_config,
    ) as (plugin, version, instance):
        with upload_file(client, "utterances.csv") as file:
            assert len(file.refresh().blocks) == 0
            file.blockify(plugin_instance=instance.handle).wait()
            # Check the number of blocks
            blocks = file.refresh().blocks
            assert len(blocks) == 5
            for block in blocks:
                assert block.tags is not None
                assert len(block.tags) > 0
                for tag in block.tags:
                    assert tag.name is not None
                    assert tag.kind is not None



from abc import ABC, abstractmethod
from typing import List, Optional

from pydantic.main import BaseModel

from steamship import Block


class LLM(BaseModel, ABC):
    """LLM wraps large language model-based backends.

    They may be used with LLMAgents in Action selection, or for direct prompt completion."""

    @abstractmethod
    def complete(self, prompt: str, stop: Optional[str] = None) -> List[Block]:
        """Completes the provided prompt, stopping when the stop sequeunce is found."""
        pass



import pytest
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.client import steamship_use_skill
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import SteamshipError
from steamship.client.steamship import SKILL_TO_PROVIDER
from steamship.data import TagValueKey


def test_use_skill():
    client = get_steamship_client()
    tagger_plugin_path = PLUGINS_PATH / "taggers" / "plugin_configurable_tagger.py"
    config_template = {
        "tagKind": {"type": "string"},
        "tagName": {"type": "string"},
        "numberValue": {"type": "number"},
        "booleanValue": {"type": "boolean"},
    }
    instance_config1 = {
        "tagKind": "testTagKind",
        "tagName": "testTagName",
        "numberValue": 5,
        "booleanValue": True,
    }

    with deploy_plugin(
        client,
        tagger_plugin_path,
        "tagger",
        version_config_template=config_template,
        instance_config=instance_config1,
    ) as (plugin, version, instance):
        plugin_handle = plugin.handle

        SKILL_TO_PROVIDER["hello"] = {
            "steamship": {
                "plugin_handle": plugin_handle,
                "config": {
                    "tagKind": "testTagKind2",
                    "tagName": "testTagName2",
                    "numberValue": 4,
                    "booleanValue": False,
                },
            }
        }

        test_str = "Hi there!"
        with pytest.raises(SteamshipError):
            with steamship_use_skill(skill="doesnotexist"):
                pass

        with pytest.raises(SteamshipError):
            with steamship_use_skill(skill="hello", provider="doesnotexist"):
                pass

        with steamship_use_skill(skill="hello", delete_workspace=False) as skill_instance_1:
            _test_skill_instance(skill_instance_1, test_str)

        with steamship_use_skill(
            skill="hello", provider="steamship", delete_workspace=False
        ) as skill_instance_1:
            _test_skill_instance(skill_instance_1, test_str)


def _test_skill_instance(skill_instance_1, test_str):
    assert skill_instance_1 is not None
    res = skill_instance_1.tag(doc=test_str)
    res.wait()
    assert res.output is not None
    assert len(res.output.file.blocks) == 1
    assert res.output.file.blocks[0].text == test_str
    # Validate configured content
    assert len(res.output.file.tags) == 1
    tag = res.output.file.tags[0]
    config = SKILL_TO_PROVIDER["hello"]["steamship"]["config"]
    assert tag.name == config["tagName"]
    assert tag.kind == config["tagKind"]
    assert tag.value[TagValueKey.NUMBER_VALUE] == config["numberValue"]
    assert tag.value[TagValueKey.BOOL_VALUE] == config["booleanValue"]



# Plugin
#
# This file contains the abstractions for managing Steamship plugins.
# To see how to implement a Steamship Plugin, see plugin_service.py in the same folder.
#
#

from __future__ import annotations

import json
from enum import Enum
from typing import Any, Dict, List, Optional, Type, Union

from pydantic import BaseModel, Field

from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import IdentifierRequest, ListRequest, Request, SortOrder, UpdateRequest
from steamship.base.response import ListResponse
from steamship.data.manifest import Manifest

from .hosting import HostingType


class CreatePluginRequest(Request):
    training_platform: Optional[HostingType] = None
    id: str = None
    type: str = None
    transport: str = None
    is_public: bool = None
    handle: str = None
    description: str = None
    metadata: str = None
    fetch_if_exists: bool = False


class PluginUpdateRequest(UpdateRequest):
    id: Optional[str] = None
    handle: Optional[str] = None
    description: Optional[str] = None
    profile: Optional[Manifest] = None
    readme: Optional[str] = None


class ListPluginsRequest(ListRequest):
    type: Optional[str] = None


class ListPluginsResponse(ListResponse):
    plugins: List[Plugin]


class PluginType(str, Enum):
    parser = "parser"
    classifier = "classifier"
    tagger = "tagger"
    embedder = "embedder"
    generator = "generator"


class PluginAdapterType(str, Enum):
    steamship_docker = "steamshipDocker"
    steamship_sagemaker = "steamshipSagemaker"
    huggingface = "huggingface"
    openai = "openai"


class PluginTargetType(str, Enum):
    FILE = "file"
    WORKSPACE = "workspace"


class Plugin(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    type: str = None
    transport: str = None
    is_public: bool = None
    training_platform: Optional[HostingType] = None
    handle: str = None
    description: str = None
    metadata: str = None
    profile: Optional[Manifest] = None
    readme: Optional[str] = None
    user_id: Optional[str] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["plugin"] if "plugin" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        description: str,
        type_: str,
        transport: str,
        is_public: bool,
        handle: str = None,
        training_platform: Optional[HostingType] = None,
        metadata: Union[str, Dict, List] = None,
        fetch_if_exists: bool = False,
    ) -> Plugin:
        if isinstance(metadata, dict) or isinstance(metadata, list):
            metadata = json.dumps(metadata)

        req = CreatePluginRequest(
            training_platform=training_platform,
            type=type_,
            transport=transport,
            is_public=is_public,
            handle=handle,
            description=description,
            metadata=metadata,
            fetch_if_exists=fetch_if_exists,
        )
        return client.post(
            "plugin/create",
            req,
            expect=Plugin,
        )

    @staticmethod
    def list(
        client: Client,
        t: str = None,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        sort_order: Optional[SortOrder] = SortOrder.DESC,
    ) -> ListPluginsResponse:
        return client.post(
            "plugin/list",
            ListPluginsRequest(
                type=t, page_size=page_size, page_token=page_token, sort_order=sort_order
            ),
            expect=ListPluginsResponse,
        )

    @staticmethod
    def get(client: Client, handle: str):
        return client.post("plugin/get", IdentifierRequest(handle=handle), expect=Plugin)

    def update(self, client: Client) -> Plugin:
        return client.post(
            "plugin/update",
            PluginUpdateRequest(
                id=self.id, description=self.description, profile=self.profile, readme=self.readme
            ),
            expect=Plugin,
        )


ListPluginsResponse.update_forward_refs()



from steamship import Block, MimeTypes
from steamship.invocable import InvocableResponse
from steamship.plugin.generator import Generator
from steamship.plugin.inputs.raw_block_and_tag_plugin_input import RawBlockAndTagPluginInput
from steamship.plugin.outputs.plugin_output import UsageReport
from steamship.plugin.outputs.raw_block_and_tag_plugin_output import RawBlockAndTagPluginOutput
from steamship.plugin.request import PluginRequest


class TestGenerator(Generator):
    def run(
        self, request: PluginRequest[RawBlockAndTagPluginInput]
    ) -> InvocableResponse[RawBlockAndTagPluginOutput]:
        image_blocks = 0
        fetched_raw = 0
        for block in request.data.blocks:
            if block.mime_type == MimeTypes.PNG:
                image_blocks += 1
                _ = block.raw()
                fetched_raw += 1

        output_text = f"Found {image_blocks} image blocks and fetched data from {fetched_raw}"
        return InvocableResponse(
            data=RawBlockAndTagPluginOutput(
                blocks=[Block(text=output_text)], usage=[UsageReport.run_units(1)]
            )
        )



"""Quickly adds a file by either downloading it or using a plugin to scrape it. Always returns a File object."""
from typing import List, Optional

import requests

from steamship.client.steamship import Steamship
from steamship.data.file import File
from steamship.data.plugin.plugin_instance import PluginInstance
from steamship.data.tags.tag import Tag


def _scrape_with_importer(client: Steamship, url: str, importer: PluginInstance) -> File:
    """Scrape a URL with the provided Importer plugin, and then await the File."""
    task = File.create_with_plugin(client, plugin_instance=importer.handle, url=url)
    task.wait()
    return task.output


def _scrape_with_download(client: Steamship, url: str) -> File:
    """Scrape the bytes of a URL and load it into a File."""
    resp = requests.get(url)
    file = File.create(
        client,
        content=resp.content,
    )
    return file


def _scrape(client: Steamship, url: str) -> File:
    """Do a best effort to scrape the provided URL, returning a File."""
    # Here we do a bit of funny business to pick the right importer.
    if "youtube.com" in url:
        importer = client.use_plugin("youtube-file-importer")
        return _scrape_with_importer(client, url, importer)
    else:
        return _scrape_with_download(client, url)


def scrape(client: Steamship, url: str, tags: Optional[List[Tag]] = None) -> File:
    """Scrape a file via URL, returning a File object no matter what."""
    file = _scrape(client, url)

    # Add the provided tags.
    for tag in tags or []:
        tag.file_id = file.id
        client.post("tag/create", tag, expect=Tag)

    return file



import itertools
import threading

import click


class Spinner(object):
    # ["   🚢", "  🚢 ", " 🚢  ", "🚢   "]
    # Unfortunately, backspacing doesn't seem to work correctly for emoji in iTerm, so leaving the "spinner"
    # as adding ships for now
    spinner_cycle = itertools.cycle(["🚢"])

    def __init__(self):
        self.stop_running = None
        self.spin_thread = None

    def start(self):
        self.stop_running = threading.Event()
        self.spin_thread = threading.Thread(target=self.init_spin)
        self.spin_thread.start()

    def stop(self):
        if self.spin_thread:
            self.stop_running.set()
            self.spin_thread.join()

    def init_spin(self):
        while not self.stop_running.is_set():
            click.echo(next(self.spinner_cycle), nl=False)
            self.stop_running.wait(1)
            # click.echo("\b", nl=False)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        return False


def ship_spinner():
    """This function creates a context manager that is used to display a
    spinner on stdout as long as the context has not exited.
    The spinner is created only if stdout is not redirected, or if the spinner
    is forced using the `force` parameter.
    """
    return Spinner()



import logging
from abc import ABC
from typing import Type

import requests
from pydantic import Field

from steamship.experimental.package_starters.web_bot import (
    SteamshipWidgetBot,
    response_for_exception,
)
from steamship.experimental.transports import TelegramTransport
from steamship.invocable import Config, InvocableResponse, post


class TelegramBotConfig(Config):
    bot_token: str = Field(description="The secret token for your Telegram bot")


class TelegramBot(SteamshipWidgetBot, ABC):
    config: TelegramBotConfig
    telegram_transport: TelegramTransport

    @classmethod
    def config_cls(cls) -> Type[Config]:
        """Return the Configuration class."""
        return TelegramBotConfig

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.api_root = f"https://api.telegram.org/bot{self.config.bot_token}"
        self.telegram_transport = TelegramTransport(
            bot_token=self.config.bot_token, client=self.client
        )  # dangerous!

    def instance_init(self):
        """This instance init method is called automatically when an instance of this package is created. It registers the URL of the instance as the Telegram webhook for messages."""
        super().instance_init()
        webhook_url = self.context.invocable_url + "respond"
        self.telegram_transport.instance_init(webhook_url=webhook_url)

    @post("respond", public=True)
    def respond(self, **kwargs) -> InvocableResponse[str]:
        """Endpoint implementing the Telegram WebHook contract. This is a PUBLIC endpoint since Telegram cannot pass a Bearer token."""

        # TODO: must reject things not from the package
        message = kwargs.get("message", {})
        chat_id = message.get("chat", {}).get("id")
        try:
            incoming_message = self.telegram_transport.parse_inbound(message)
            if incoming_message is not None:
                response = self.create_response(incoming_message)
                if response is not None:
                    self.telegram_transport.send(response)
                else:
                    # Do nothing here; this could be a message we intentionally don't want to respond to (ex. an image or file upload)
                    pass
            else:
                # Do nothing here; this could be a message we intentionally don't want to respond to (ex. an image or file upload)
                pass
        except Exception as e:
            response = response_for_exception(e, chat_id=chat_id)

            if chat_id is not None:
                self.telegram_transport.send([response])
        # Even if we do nothing, make sure we return ok
        return InvocableResponse(string="OK")

    @post("webhook_info")
    def webhook_info(self) -> dict:
        return requests.get(self.api_root + "/getWebhookInfo").json()

    @post("info")
    def info(self) -> dict:
        """Endpoint returning information about this bot."""
        return self.telegram_transport.info()

    @post("disconnect_webhook")
    def disconnect_webhook(self) -> InvocableResponse[str]:
        self.telegram_transport.instance_deinit()
        logging.info("Disconnected webhook.")
        return InvocableResponse(data="OK")



from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from .base import (
    Configuration,
    MimeTypes,
    RuntimeEnvironments,
    SteamshipError,
    Task,
    TaskState,
    check_environment,
)
from .data import (
    Block,
    DocTag,
    EmbeddingIndex,
    File,
    Package,
    PackageInstance,
    PackageVersion,
    PluginInstance,
    PluginVersion,
    Tag,
    Workspace,
)

from .client import Steamship  # isort:skip

__all__ = [
    "Steamship",
    "Configuration",
    "SteamshipError",
    "MimeTypes",
    "Package",
    "PackageInstance",
    "PackageVersion",
    "File",
    "Task",
    "TaskState",
    "Block",
    "Tag",
    "Workspace",
    "PluginInstance",
    "PluginVersion",
    "DocTag",
    "EmbeddingIndex",
    "check_environment",
    "RuntimeEnvironments",
]



from abc import ABC, abstractmethod
from typing import List, Optional

from steamship import Block
from steamship.experimental.transports.steamship_widget import SteamshipWidgetTransport
from steamship.invocable import PackageService, post


def response_for_exception(e: Optional[Exception], chat_id: Optional[str] = None) -> Block:
    return_text = f"An error happened while creating a response: {e}"
    if e is None:
        return_text = "An unknown error happened. Please reach out to support@steamship.com or on our discord at https://steamship.com/discord"

    if "usage limit" in f"{e}":
        return_text = "You have reached the introductory limit of Steamship. Visit https://steamship.com/account/plan to sign up for a plan."

    result = Block(text=return_text)
    result.set_chat_id(chat_id)
    return result


class SteamshipWidgetBot(PackageService, ABC):
    steamship_widget_transport: SteamshipWidgetTransport

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.steamship_widget_transport = SteamshipWidgetTransport(client=self.client)

    @abstractmethod
    def create_response(self, incoming_message: Block) -> Optional[List[Block]]:
        raise NotImplementedError

    @post("answer", public=True)
    def answer(self, **payload) -> List[Block]:
        """Endpoint that implements the contract for Steamship embeddable chat widgets. This is a PUBLIC endpoint since these webhooks do not pass a token."""
        incoming_message = self.steamship_widget_transport.parse_inbound(payload)
        try:
            response = self.create_response(incoming_message)
        except Exception as e:
            response = response_for_exception(e, chat_id=incoming_message.chat_id)

        # We don't call self.steamship_widget_transport.send because the result is the return value
        return response



import logging
from typing import List

import tiktoken

from steamship import Block, SteamshipError
from steamship.data.tags.tag_constants import RoleTag


def token_length(block: Block, tiktoken_encoder: str = "p50k_base") -> int:
    """Calculate num tokens with tiktoken package."""
    # create a GPT-3 encoder instance
    enc = tiktoken.get_encoding(tiktoken_encoder)
    # encode the text using the GPT-3 encoder
    tokenized_text = enc.encode(block.text)
    # calculate the number of tokens in the encoded text
    return len(tokenized_text)


# TODO: abstract this into one of N possible strategies
def filter_blocks_for_prompt_length(max_tokens: int, blocks: List[Block]) -> List[int]:

    retained_blocks = []
    total_length = 0

    # Keep all system blocks
    for block in blocks:
        if block.chat_role == RoleTag.SYSTEM:
            retained_blocks.append(block)
            total_length += token_length(block)

    # If system blocks are too long, throw error
    if total_length > max_tokens:
        raise SteamshipError(
            f"Plugin attempted to filter input to fit into {max_tokens} tokens, but the total size of system blocks was {total_length}"
        )

    # Now work backwards and keep as many blocks as we can
    num_system_blocks = len(retained_blocks)
    for block in reversed(blocks):
        if block.chat_role != RoleTag.SYSTEM and total_length < max_tokens:
            block_length = token_length(block)
            if block_length + total_length < max_tokens:
                retained_blocks.append(block)
                total_length += block_length
                logging.info(f"Adding block {block.index_in_file} of token length {block_length}")

    # If we didn't add any non-system blocks, throw error
    if len(retained_blocks) == num_system_blocks:
        raise SteamshipError(
            f"Plugin attempted to filter input to fit into {max_tokens} tokens, but no non-System blocks remained."
        )

    block_indices = [block.index_in_file for block in blocks if block in retained_blocks]
    logging.info(f"Filtered input.  Total tokens {total_length} Block indices: {block_indices}")
    return block_indices



from typing import Any, List, Union

from steamship import Block, Steamship, Task
from steamship.agents.llms import OpenAI
from steamship.agents.schema import AgentContext, Tool
from steamship.agents.utils import get_llm, with_llm
from steamship.utils.repl import ToolREPL

DEFAULT_PROMPT = """Instructions:
Please rewrite the following passage to be incredibly polite, to a fault.
Passage:
{input}
Rewritten Passage:"""


class TextRewritingTool(Tool):
    """
    Example tool to illustrate rewriting a statement according to a particular personality.
    """

    rewrite_prompt: str = DEFAULT_PROMPT
    name: str = "TextRewritingTool"
    human_description: str = "Rewrites a piece of text using the provided prompt."
    agent_description: str = "Used to rewrite a piece of text given a prompt. Takes text as input, and provides text as output."

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        """Rewrites each provided text block using the stored prompt. Non-text blocks are passed through without
        modification.

        Inputs
        ------
        input: List[Block]
            A list of blocks to be rewritten if they contain text. Each block will be considered a separate input.
        memory: AgentContext
            The active AgentContext.

        Output
        ------
        output: List[Blocks]
            A list of blocks whose content has been rewritten. Synchronously produced (for now).
        """
        llm = get_llm(context)

        blocks = []
        for block in tool_input:
            # If the block is not text, simply pass it through.
            if not block.is_text():
                continue

            # If the block is text, rewrite it and append that output.
            prompt = self.rewrite_prompt.format(input=block.text)
            output_blocks = llm.complete(prompt=prompt)
            blocks.extend(output_blocks)

        return blocks


if __name__ == "__main__":
    with Steamship.temporary_workspace() as client:
        ToolREPL(TextRewritingTool()).run_with_client(
            client=client, context=with_llm(llm=OpenAI(client=client))
        )



from abc import ABC, abstractmethod

from pydantic import BaseModel

from steamship.agents.schema.action import Action
from steamship.agents.schema.context import AgentContext


class OutputParser(BaseModel, ABC):
    """Used to convert text into Actions.

    Primarily used by LLM-based agents that generate textual descriptions of
    selected actions and their inputs. OutputParsers can be used to convert
    those descriptions into Action objects for the AgentService to run.

    Example:
     - input: "Action: GenerateImage
               ActionInput: row-house"
     - output: Action("dalle", "row-house")
    """

    @abstractmethod
    def parse(self, text: str, context: AgentContext) -> Action:
        """Convert text into an Action object."""
        pass



import uuid
from typing import Optional

from steamship import Block, SteamshipError
from steamship.agents.schema import Metadata
from steamship.experimental.transports.transport import Transport

API_BASE = "https://api.telegram.org/bot"


class SteamshipWidgetTransport(Transport):
    """Experimental base class to encapsulate a Steamship web widget communication channel."""

    def _instance_init(self, *args, **kwargs):
        pass

    def _instance_deinit(self, *args, **kwargs):
        """Unsubscribe from updates."""
        pass

    def _send(self, blocks: [Block], metadata: Metadata):
        """Send a response to the client.

        TODO: Since this isn't a push, but rather an API return, we need to figure out how to model this.
        """
        pass

    def _info(self) -> dict:
        """Fetches info about this bot."""
        return {}

    def _parse_inbound(self, payload: dict, context: Optional[dict] = None) -> Optional[Block]:
        """Parses an inbound Steamship widget message."""

        message_text = payload.get("question")
        if message_text is None:
            raise SteamshipError(f"No 'question' found in Steamship widget message: {payload}")

        chat_id = payload.get("chat_session_id", "default")

        message_id = str(uuid.uuid4())

        result = Block(text=message_text)
        result.set_chat_id(str(chat_id))
        result.set_message_id(str(message_id))
        return result



from typing import List, Optional

from steamship.agents.tools.text_generation.text_rewrite_tool import TextRewritingTool
from steamship.utils.repl import ToolREPL

DEFAULT_FACTS = [
    "New York City has 424 subway stations",
    "Boston has 51 subway stations",
    "Washington, DC has 97 subway stations",
    "San Francisco has 47 subway stations",
    "Miami has 23 subway stations",
    "Los Angeles has 16 subway stations",
    "Chicago has 145 subway stations",
    "Baltimore has 14 subway stations",
]

DEFAULT_QUESTION_ANSWERING_PROMPT = (
    "Use the following pieces of memory to answer the question at the end. "
    """If you don't know the answer, just say that you don't know, don't try to make up an answer.

{source_text}

Question: {{input}}

Helpful Answer:"""
)


class PromptDatabaseQATool(TextRewritingTool):
    """
    Example tool to illustrate how one can create a tool with a mini database embedded in a prompt.

    To use:

        tool = PromptDatabaseQATool(
            facts=[
                "Sentence with fact 1",
                "Sentence with fact 2"
            ],
            ai_description="Used to answer questions about SPECIFIC_THING. "
                           "The input is the question and the output is the answer."
        )

    """  # noqa: RST203, RST301

    name = "PromptDatabaseQATool"
    human_description = "Answers questions about the number of subway stations in US cities."
    agent_description = "Used to answer questions about the number of subway stations in US cities. The input is the question about subway stations. The output is the answer as a sentence."
    question_answering_prompt: Optional[str] = DEFAULT_QUESTION_ANSWERING_PROMPT
    facts: List[str] = DEFAULT_FACTS

    def __init__(self, facts: Optional[List[str]] = None, **kwargs):
        _rewrite_prompt = kwargs.get("question_answering_prompt", DEFAULT_QUESTION_ANSWERING_PROMPT)
        _fact_list = [f"- {fact}" for fact in facts or DEFAULT_FACTS]

        kwargs["rewrite_prompt"] = kwargs.get(
            "rewrite_prompt", _rewrite_prompt.format(source_text="\n".join(_fact_list))
        )
        super().__init__(**kwargs)


if __name__ == "__main__":
    ToolREPL(PromptDatabaseQATool()).run()



"""CSV Blockifier - Steamship Plugin."""

from typing import Any, Dict

from assets.plugins.blockifiers.csv_blockifier import CsvBlockifier

from steamship.invocable import InvocationContext, create_handler
from steamship.plugin.blockifier.blockifier import Blockifier


class TsvBlockifier(CsvBlockifier, Blockifier):
    """Converts TSV into Tagged Steamship Blocks.

    Implementation is only here to demonstrate how plugins can be built through inheritance.
    """

    def __init__(
        self, client=None, config: Dict[str, Any] = None, context: InvocationContext = None
    ):
        super().__init__(client, config, context)
        self.config.delimiter = "\t"


handler = create_handler(TsvBlockifier)



import re
from typing import Any, List, Union

import requests

from steamship import Block, Task
from steamship.agents.schema import AgentContext, Tool
from steamship.utils.repl import ToolREPL


class FetchAudioUrlsFromRssTool(Tool):
    """Given an RSS feed, this tool will extract episode URLs."""

    name: str = "FetchAudioUrlsFromRssTool"
    human_description: str = "Fetches the episode URLs from a Podcast RSS feed."
    agent_description: str = (
        "Used to fetch the podcast episode URLs from a podcast RSS feed. "
        "The input is the URL of the RSS feed. "
        "The output is the URLs of the episode audio."
    )

    def _get_audio_urls(self, url: str, context: AgentContext) -> List[str]:
        response = requests.get(url)
        pattern = re.compile(r"<enclosure[^>]+url\s*=\s*\"([^\"]+)\"", re.IGNORECASE)
        urls = []
        for match in pattern.finditer(response.text):
            urls.append(match.group(1))
        return urls

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        blocks = []
        for input_block in tool_input:
            if not input_block.is_text():
                continue
            url = input_block.text
            urls = self._get_audio_urls(url, context)
            blocks.extend([Block(text=url) for url in urls])
        return blocks


if __name__ == "__main__":
    ToolREPL(FetchAudioUrlsFromRssTool()).run()

# Try with https://anchor.fm/s/e1369b4c/podcast/rss



from steamship import File, Tag
from steamship.invocable import Invocable, InvocableResponse, post


class PackageWithInstanceInit(Invocable):
    def instance_init(self):
        """Creates a file in the workspace to indicate that it's been called"""
        File.create(
            client=self.client, content="", tags=[Tag(kind="test", name=self.context.invocable_url)]
        )

    @post("was_init_called")
    def was_init_called(self) -> InvocableResponse:
        """Verifies that init was called by finding the created file"""
        for file in File.list(self.client).files:
            for tag in file.tags:
                if tag.kind == "test":
                    return InvocableResponse(data=tag.name)
        return InvocableResponse(data="NOPE")



from steamship import Block, SteamshipError, Task, TaskState
from steamship.invocable import InvocableResponse
from steamship.plugin.generator import Generator
from steamship.plugin.inputs.raw_block_and_tag_plugin_input import RawBlockAndTagPluginInput
from steamship.plugin.outputs.plugin_output import UsageReport
from steamship.plugin.outputs.raw_block_and_tag_plugin_output import RawBlockAndTagPluginOutput
from steamship.plugin.request import PluginRequest

STATUS_MESSAGE = "I'm an asynchronous generator!"
ASYNC_JOB_ID = "JOB-ID"
OUTPUT_BLOCK_TEXT = "Look at me, I did this asynchronously!"


class TestGenerator(Generator):
    def run(
        self, request: PluginRequest[RawBlockAndTagPluginInput]
    ) -> InvocableResponse[RawBlockAndTagPluginOutput]:

        if request.is_status_check:
            if request.status.remote_status_input.get("async_job_id") != ASYNC_JOB_ID:
                raise SteamshipError(message="Required status password was not provided")

            result = [Block(text=OUTPUT_BLOCK_TEXT)]

            return InvocableResponse(
                data=RawBlockAndTagPluginOutput(blocks=result, usage=[UsageReport.run_tokens(5)])
            )
        else:
            return InvocableResponse(
                status=Task(
                    state=TaskState.running,
                    remote_status_message=STATUS_MESSAGE,
                    remote_status_input={
                        "async_job_id": ASYNC_JOB_ID,
                        "data": {},
                    },
                )
            )



# noinspection PyPackageRequirements
import uuid

import pytest
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import PluginInstance, SteamshipError, Workspace
from steamship.data.plugin import Plugin, PluginAdapterType, PluginType
from steamship.data.user import User


def test_plugin_create():
    steamship = get_steamship_client()

    resp = Plugin.list(steamship)
    prior_plugin_ids = [plugin.id for plugin in resp.plugins]

    plugin_args = {
        "client": steamship,
        "description": "This is just for test",
        "type_": PluginType.tagger,
        "transport": PluginAdapterType.steamship_docker,
        "is_public": True,
    }

    required_fields = {"description", "type_"}
    for required_field in required_fields:
        del plugin_args[required_field]
        with pytest.raises(TypeError):
            Plugin.create(**plugin_args)

    plugin = Plugin.create(
        client=steamship,
        description="This is just for test",
        type_=PluginType.tagger,
        transport=PluginAdapterType.steamship_docker,
        is_public=False,
    )

    assert plugin.id not in prior_plugin_ids

    resp = Plugin.list(steamship)
    assert plugin.id in [
        plugin.id for plugin in resp.plugins
    ]  # newly-created plugin should be top of list
    assert plugin.description in [plugin.description for plugin in resp.plugins]

    user_id = User.current(steamship).id
    assert plugin.user_id == user_id

    # No fetch_if_exists doesn't work
    with pytest.raises(SteamshipError):
        _ = Plugin.create(
            client=steamship,
            handle=plugin.handle,
            description="This is just for test",
            type_=PluginType.tagger,
            transport=PluginAdapterType.steamship_docker,
            is_public=False,
        )


def test_plugin_public():
    steamship = get_steamship_client()

    resp = Plugin.list(steamship)
    assert resp.plugins is not None
    plugins = resp.plugins

    assert len(plugins) > 0


def test_deploy_in_workspace():
    client = get_steamship_client()
    workspace = Workspace.create(client, handle="test-non-default-workspace")
    client.switch_workspace(workspace_id=workspace.id)
    instance = PluginInstance.create(client, plugin_handle="test-tagger")
    assert instance.workspace_id == workspace.id


def test_plugin_instance_get():
    steamship = get_steamship_client()
    handle = f"test_tagger_test_handle{uuid.uuid4()}"
    instance = PluginInstance.create(steamship, plugin_handle="test-tagger", handle=handle)
    assert instance.id is not None

    other_instance = PluginInstance.get(steamship, handle=handle)
    assert instance.id == other_instance.id


def test_plugin_update():
    steamship = get_steamship_client()

    plugin = Plugin.create(
        client=steamship,
        description="This is just for test",
        type_=PluginType.tagger,
        transport=PluginAdapterType.steamship_docker,
        is_public=False,
    )

    new_description = "f"
    plugin.description = new_description

    plugin = plugin.update(steamship)
    assert plugin.description == new_description

    got_plugin = Plugin.get(steamship, handle=plugin.handle)
    assert got_plugin.description == new_description


def test_plugin_instance_quick_create():
    steamship = get_steamship_client()

    handle = f"test_tagger_test_handle{uuid.uuid4()}"
    handle2 = f"test_tagger_test_handle{uuid.uuid4()}"

    p1 = steamship.use_plugin("test-tagger", handle)
    p2 = steamship.use_plugin("test-tagger", handle2)
    p3 = steamship.use_plugin("test-tagger", handle)
    p4 = steamship.use_plugin("test-tagger")

    assert p1.id == p3.id
    assert p1.id != p2.id
    assert p1.id != p4.id


def test_plugin_instance_handle_refs():
    steamship = get_steamship_client()
    handle = f"test_tagger_test_handle{uuid.uuid4()}"
    instance = PluginInstance.create(
        steamship, plugin_handle="test-tagger", plugin_version_handle="1.0", handle=handle
    )

    assert instance.plugin_handle == "test-tagger"
    assert instance.plugin_version_handle == "1.0"

    got_instance = PluginInstance.get(steamship, handle=handle)
    assert got_instance.plugin_handle == "test-tagger"
    assert got_instance.plugin_version_handle == "1.0"

    use_instance = steamship.use_plugin("test-tagger", handle)
    assert use_instance.plugin_handle == "test-tagger"
    assert use_instance.plugin_version_handle == "1.0"



from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import zip_deployable
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import Package, PackageVersion


def test_version_create():
    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "demo_package.py"

    package = Package.create(client)
    zip_bytes = zip_deployable(demo_package_path)

    version = PackageVersion.create(client, package_id=package.id, filebytes=zip_bytes)

    assert version is not None



import threading
from socketserver import TCPServer
from typing import Optional, Type

from steamship.cli.local_server.handler import make_handler
from steamship.invocable import Invocable


class SteamshipHTTPServer:
    """A simple HTTP Server that wraps an invocable (package or plugin).

    To use, call:

       server = SteamshipHTTPServer(invocable)
       server.start()

    To shut down, call:

       server.stop()

    """

    invocable: Type[Invocable]
    port: int
    server: TCPServer
    default_api_key: Optional[str] = (None,)
    invocable_handle: str = (None,)
    invocable_version_handle: str = (None,)
    invocable_instance_handle: str = (None,)

    def __init__(
        self,
        invocable: Type[Invocable],
        port: int = 8080,
        default_api_key: Optional[str] = None,
        invocable_handle: str = None,
        invocable_version_handle: str = None,
        invocable_instance_handle: str = None,
    ):
        self.invocable = invocable
        self.port = port
        self.default_api_key = default_api_key
        self.invocable_handle = invocable_handle
        self.invocable_version_handle = invocable_version_handle
        self.invocable_instance_handle = invocable_instance_handle

    def start(self):
        """Start the server."""
        self.server = TCPServer(
            ("", self.port),
            make_handler(
                self.invocable,
                self.port,
                default_api_key=self.default_api_key,
                invocable_handle=self.invocable_handle,
                invocable_version_handle=self.invocable_version_handle,
                invocable_instance_handle=self.invocable_instance_handle,
            ),
        )
        self.server.serve_forever()

    def stop(self):
        """Stop the server.

        Note: This has to be called from a different thread or else it will deadlock.
        """
        _server = self.server

        class ShutdownThread(threading.Thread):
            def __init__(self):
                threading.Thread.__init__(self)

            def run(self):
                nonlocal _server
                _server.shutdown()

        shutdown_thread = ShutdownThread()
        shutdown_thread.start()



from .image_prompt_generator_tool import ImagePromptGeneratorTool
from .json_object_generator import JsonObjectGeneratorTool
from .personality_tool import PersonalityTool
from .summarize_text_with_prompt_tool import SummarizeTextWithPromptTool
from .text_rewrite_tool import TextRewritingTool
from .text_translation_tool import TextTranslationTool

__all__ = [
    "ImagePromptGeneratorTool",
    "PersonalityTool",
    "SummarizeTextWithPromptTool",
    "TextRewritingTool",
    "TextTranslationTool",
    "JsonObjectGeneratorTool",
]



import pytest

from steamship import Steamship
from steamship.agents.schema import AgentContext


@pytest.mark.usefixtures("client")
def test_same_context(client: Steamship):
    context1 = AgentContext.get_or_create(client=client, context_keys={"one": "one", "two": "two"})
    contextsame = AgentContext.get_or_create(
        client=client, context_keys={"one": "one", "two": "two"}
    )

    assert context1.id == contextsame.id


@pytest.mark.usefixtures("client")
def test_different_context(client: Steamship):
    context1 = AgentContext.get_or_create(client=client, context_keys={"one": "one", "two": "two"})
    contextdiff = AgentContext.get_or_create(
        client=client, context_keys={"one": "one", "two": "three"}
    )
    assert context1.id != contextdiff.id



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar, Union

# Note!
# =====
#
# This the files in this package are for Plugin Implementors.
# If you are using the Steamship Client, you probably are looking for either steamship.client or steamship.data
#
from steamship.invocable import Invocable, InvocableResponse
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import TrainingParameterPluginOutput
from steamship.plugin.request import PluginRequest
from steamship.plugin.trainable_model import TrainableModel

IN = TypeVar("IN")
OUT = TypeVar("OUT")


class PluginService(Invocable, Generic[IN, OUT], ABC):
    """The Abstract Base Class of a Steamship Plugin.

    All Steamship Plugins implement the operation:

    - run(PluginRequest[T]) -> Response[U]

    Many plugins are effectively stateless. This run operation defines their entire capability.
    Examples of such stateless plugins are:
    - File Import Plugin
    - Export Plugin

    Other plugins have state but in a very controlled way:
    - they can be trained,
    - this trainable process produces a "model",
    - that model acts as the state on which the `run` method is conditioned

    This model is stored in the Steamship Workspace that owns the Plugin Instance, and access to it is provided by the
    hosting environment that runs the model.
    - TODO(ted) Document this process.

    These stateful plugins are called "Trainable Plugins," and they must implement the following additional methods:

    - get_training_parameters(PluginRequest[TrainingParameterInput]) -> Response[TrainingParameterOutput]
    - train(PluginRequest[TrainPluginInput]) -> Response[TrainPluginOutput]

    """

    @abstractmethod
    def run(self, request: PluginRequest[IN]) -> Union[OUT, InvocableResponse[OUT]]:
        """Runs the core operation implemented by this plugin: import, export, blockify, tag, etc.

        This is the method that a Steamship Plugin implements to perform its main work.
        """
        pass


class TrainablePluginService(PluginService, Generic[IN, OUT], ABC):
    @abstractmethod
    def model_cls(self) -> Type[TrainableModel]:
        """Returns the constructor of the TrainableModel this TrainablePluginService uses.

        This is required so the `run` method below can load the model and provide it to the subclass implementor.
        """
        pass

    def run(self, request: PluginRequest[IN]) -> Union[OUT, InvocableResponse[OUT]]:
        """Loads the trainable model before passing the request to the `run_with_model` handler on the subclass."""
        logging.info("TrainablePluginService:run() - Loading model")
        model = self.model_cls().load_remote(
            client=self.client,  # This field comes from being a subclass of App
            plugin_instance_id=request.context.plugin_instance_id,
            checkpoint_handle=None,  # Will use default
            use_cache=True,
            plugin_instance_config=self.config,
        )
        logging.info("TrainablePluginService:run() - Loaded model; invoking run_with_model")
        return self.run_with_model(request, model)

    @abstractmethod
    def run_with_model(
        self, request: PluginRequest[IN], model: TrainableModel
    ) -> Union[OUT, InvocableResponse[OUT]]:
        """Rather than implementing run(request), a TrainablePluginService implements run_with_model(request, model)"""
        pass

    @abstractmethod
    def get_training_parameters(
        self, request: PluginRequest[TrainingParameterPluginInput]
    ) -> InvocableResponse[TrainingParameterPluginOutput]:
        """Produces the trainable parameters for this plugin.

        This method is run by the Steamship Engine prior to training to fetch hyperparameters.

        - The user themselves can provide hyperparameters on the TrainingParameterPluginInput object.
        - This method then transforms those into the TrainingParameterPluginOutput object, altering the user's values
          if desired.
        - The Engine then takes those TrainingParameterPluginOutput and presents them on the TrainPluginInput

        """
        pass

    @abstractmethod
    def train(
        self, request: PluginRequest[TrainPluginInput], model: TrainableModel
    ) -> InvocableResponse[TrainPluginOutput]:
        """Train the model."""
        pass

    @abstractmethod
    def train_status(
        self, request: PluginRequest[TrainPluginInput], model: TrainableModel
    ) -> InvocableResponse[TrainPluginOutput]:
        """Train the model."""
        pass



import pytest
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.client import register_plugin_instance_subclass, steamship_use_plugin
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import SteamshipError
from steamship.data.plugin.prompt_generation_plugin_instance import PromptGenerationPluginInstance


def test_use_prompt():

    client = get_steamship_client()
    tagger_plugin_path = PLUGINS_PATH / "taggers" / "plugin_prompt_generator.py"

    with deploy_plugin(
        client,
        tagger_plugin_path,
        "tagger",
    ) as (plugin, version, instance):
        # Add this plugin to the list which will be cast into the PromptCompletionPluginInstance class.
        register_plugin_instance_subclass(plugin.handle, PromptGenerationPluginInstance)

        # Now we just invoke use_plugin.
        # Note: this just applies Steamship.use_plugin with the same testing environment as the unit tests.
        with steamship_use_plugin(plugin.handle, delete_workspace=True) as llm:
            # And we can run a prompt. Covering some bases to encode edge cases and their expected behavior.

            # Prompt vars: 1, Provided vars: 1
            assert llm.generate("Hello, {name}!", {"name": "world"}) == "Hello, world!"
            # Prompt vars: 0, Provided vars: 1
            assert llm.generate("Hello, World!", {"name": "world"}) == "Hello, World!"
            # Prompt vars: 1, Provided vars: 0
            with pytest.raises(SteamshipError):
                assert llm.generate("Hello, {name}!") == "Hello, {name}!"
            # Prompt vars: 0, Provided vars: 0
            assert llm.generate("Hello, World!") == "Hello, World!"



from .base_tools import (
    AudioBlockifierTool,
    AudioGeneratorTool,
    GeneratorTool,
    ImageBlockifierTool,
    ImageGeneratorTool,
    ScrapeAndBlockifyTool,
)

__all__ = [
    "AudioGeneratorTool",
    "AudioBlockifierTool",
    "ImageBlockifierTool",
    "ImageGeneratorTool",
    "GeneratorTool",
    "ScrapeAndBlockifyTool",
]



class longstr(str):  # noqa: N801
    """Long string functions mostly as a type annotation for the web."""

    pass


class fileurl(str):  # noqa: N801
    """Type alias that, if used in a package method argument, will cause a file upload widget to appear."""

    pass



from steamship.plugin.inputs.train_plugin_input import TrainPluginInput


def test_training_input():
    input = TrainPluginInput.parse_obj({"pluginInstance": "test", "trainingDataUrl": "test"})
    assert input.training_data_url == "test"



from steamship_tests.base.test_task import NoOpResult

from steamship.base.tasks import Task, TaskState

PLUGIN_INSTANCE_ID = "0000-0000-0000-0000"


def create_dummy_training_task(client) -> Task[NoOpResult]:
    # Create a task that is running in the background
    task = client.post("task/noop", expect=NoOpResult, as_background_task=True)
    assert task is not None
    assert task.state == TaskState.waiting

    # Allow it to complete
    task.wait()
    assert task.state == TaskState.succeeded

    return task



from assets.packages.package_verifying_instance_init import PackageWithInstanceInit
from assets.plugins.generators.plugin_with_instance_init import TestGeneratorWithInit
from assets.plugins.taggers.plugin_trainable_tagger import TestTrainableTaggerPlugin
from steamship_tests import PACKAGES_PATH, PLUGINS_PATH
from steamship_tests.utils.client import TESTING_PROFILE
from steamship_tests.utils.deployables import deploy_package, deploy_plugin

from steamship import Steamship
from steamship.data.invocable_init_status import InvocableInitStatus


def test_instance_init_is_called_on_package():
    demo_package_path = PACKAGES_PATH / "package_verifying_instance_init.py"

    with Steamship.temporary_workspace(profile=TESTING_PROFILE) as client:
        with deploy_package(client, demo_package_path, wait_for_init=False) as (_, _, instance):
            assert instance.init_status == InvocableInitStatus.INITIALIZING
            instance.wait_for_init()
            assert instance.init_status == InvocableInitStatus.COMPLETE
            invocation_url = instance.invoke("was_init_called")
            assert invocation_url == instance.invocation_url


def test_plugin_init_dir():
    plugin = TestGeneratorWithInit()
    dir = plugin.__steamship_dir__()
    assert len(dir.get("methods", [])) == 4


def test_package_init_dir():
    package = PackageWithInstanceInit()
    dir = package.__steamship_dir__()
    assert len(dir.get("methods", [])) == 4


def test_instance_init_is_called_on_plugin():
    generator_with_init_path = PLUGINS_PATH / "generators" / "plugin_with_instance_init.py"

    with Steamship.temporary_workspace(profile=TESTING_PROFILE) as client:
        with deploy_plugin(client, generator_with_init_path, "generator", wait_for_init=False) as (
            _,
            _,
            instance,
        ):
            assert instance.init_status == InvocableInitStatus.INITIALIZING
            instance.wait_for_init()
            assert instance.init_status == InvocableInitStatus.COMPLETE

            generate_task = instance.generate(text="Yo! Banana boy!")
            generate_task.wait()
            assert generate_task.output.blocks[0].text == "!yob ananaB !oY"


def test_instance_init_on_trainable_plugin():
    plugin = TestTrainableTaggerPlugin()
    response = plugin.invocable_instance_init()
    assert response.data



import uuid

import pytest

from steamship import DocTag
from steamship.agents.memory import ChatHistory
from steamship.client import Steamship
from steamship.data import TagKind
from steamship.data.tags.tag_constants import ChatTag, RoleTag, TagValueKey


@pytest.mark.usefixtures("client")
def test_chat_create(client: Steamship):
    chat = ChatHistory.get_or_create(client, context_keys={"test_id": uuid.uuid4().hex})

    assert chat.client is not None
    assert isinstance(chat, ChatHistory)
    assert len(chat.tags) == 2
    assert chat.tags[0].kind == TagKind.DOCUMENT
    assert chat.tags[0].name == DocTag.CHAT


@pytest.mark.usefixtures("client")
def test_chat_append_system(client: Steamship):
    chat = ChatHistory.get_or_create(client, context_keys={"test_id": uuid.uuid4().hex})

    chat.append_system_message(text="some system text")
    chat.refresh()

    assert len(chat.messages) == 1
    assert chat.messages[0].text == "some system text"

    assert len(chat.messages[0].tags) == 1
    assert chat.messages[0].tags[0].kind == TagKind.CHAT
    assert chat.messages[0].tags[0].name == ChatTag.ROLE
    assert chat.messages[0].tags[0].value == {TagValueKey.STRING_VALUE: RoleTag.SYSTEM}


@pytest.mark.usefixtures("client")
def test_chat_append_user(client: Steamship):
    chat = ChatHistory.get_or_create(client, context_keys={"test_id": uuid.uuid4().hex})

    chat.append_user_message(text="some user text")
    chat.refresh()

    assert len(chat.messages) == 1
    assert chat.messages[0].text == "some user text"

    assert len(chat.messages[0].tags) == 1
    assert chat.messages[0].tags[0].kind == TagKind.CHAT
    assert chat.messages[0].tags[0].name == ChatTag.ROLE
    assert chat.messages[0].tags[0].value == {TagValueKey.STRING_VALUE: RoleTag.USER}



import pytest
from assets.packages.multi_inheritance import MultiInheritanceBot

from steamship import Steamship
from steamship.invocable import PackageService


@pytest.mark.usefixtures("client")
def test_multi_inheritance_dir(client: Steamship):

    bot = MultiInheritanceBot(client=client, config={"botToken": "invalid-bot-token"})
    bot_dir = bot.__steamship_dir__()

    methods_by_path = {method["path"]: method for method in bot_dir["methods"]}

    # assert "/take_action" in methods_by_path.keys()  # A method from AgentService
    assert "/disconnect_webhook" in methods_by_path.keys()  # A method from TelegramBot


@pytest.mark.usefixtures("client")
def test_multi_inheritance_instance_init(client: Steamship):
    class PackageServiceA(PackageService):
        a = 0

        def instance_init(self):
            self.a += 1

    class PackageServiceB(PackageService):
        b = 0

        def instance_init(self):
            self.b += 1

    # If the child explicitly calls both super instance_inits, they get called
    class PackageServiceC(PackageServiceA, PackageServiceB):
        def instance_init(self):
            PackageServiceA.instance_init(self)
            PackageServiceB.instance_init(self)

    child = PackageServiceC(client=client)

    # Does this call all superclass instance inits?
    child.instance_init()

    assert child.a == 1
    assert child.b == 1

    # Does this call all superclass instance inits?
    child.invocable_instance_init()

    assert child.a == 2
    assert child.b == 2


@pytest.mark.usefixtures("client")
def test_multi_inheritance_first_class_instance_init(client: Steamship):
    class PackageServiceA(PackageService):
        a = 0

        def instance_init(self):
            self.a += 1

    class PackageServiceB(PackageService):
        b = 0

        def instance_init(self):
            self.b += 1

    # If the child explicitly calls both super instance_inits, they get called
    class PackageServiceC(PackageServiceA, PackageServiceB):
        pass

    child = PackageServiceC(client=client)

    child.invocable_instance_init()

    assert child.a == 1
    assert child.b == 0



from .tag import Tag
from .tag_constants import DocTag, GenerationTag, TagKind, TagValueKey, TokenTag

__all__ = [
    "DocTag",
    "Tag",
    "TagKind",
    "TokenTag",
    "TagValueKey",
    "GenerationTag",
]



from typing import Optional

import pytest
from steamship_tests import TEST_ASSETS_PATH

from steamship import SteamshipError
from steamship.invocable.config import Config


class EmptyConfig(Config):
    pass


class NonEmptyConfig(Config):
    string_a: Optional[str]
    string_b: Optional[str]
    int_a: Optional[int]
    int_b: Optional[int]


TEST_CONFIGS_PATH = TEST_ASSETS_PATH / "configs"
EMPTY_JSON = TEST_CONFIGS_PATH / "empty_json.json"
ONE_ONE = TEST_CONFIGS_PATH / "one_string_one_int.json"
SINGLE_INTEGER = TEST_CONFIGS_PATH / "single_integer.json"
NOT_EXIST = TEST_CONFIGS_PATH / "__doesnt_exist__.json"

NON_EMPTY_1 = NonEmptyConfig(
    string_a="string_a",
    string_b="string_b",
    int_a=5,
    int_b=6,
)
NON_EMPTY_2 = NonEmptyConfig(
    string_a="string_a",
    int_a=5,
)
NON_EMPTY_3 = NonEmptyConfig(
    string_b="string_b",
    int_b=6,
)


# Tests plugin initialization
# --------------------------------------------


def test_json_load_into_empty_config():
    c = EmptyConfig()
    c.extend_with_json_file(EMPTY_JSON)
    c.extend_with_json_file(EMPTY_JSON, overwrite=True)
    c.extend_with_json_file(ONE_ONE)
    c.extend_with_json_file(ONE_ONE, overwrite=True)
    c.extend_with_json_file(NOT_EXIST, fail_on_missing_file=False)


def test_error_conditions():
    c = EmptyConfig()

    with pytest.raises(SteamshipError):
        c.extend_with_json_file(NOT_EXIST, fail_on_missing_file=True)

    with pytest.raises(SteamshipError):
        # Default is fail on missing
        c.extend_with_json_file(NOT_EXIST)

    with pytest.raises(SteamshipError):
        # Bad parse
        c.extend_with_json_file(SINGLE_INTEGER)


def test_no_override():
    c1 = NON_EMPTY_1.copy()
    c1.extend_with_json_file(ONE_ONE)
    assert c1.int_a == NON_EMPTY_1.int_a
    assert c1.int_b == NON_EMPTY_1.int_b
    assert c1.string_a == NON_EMPTY_1.string_a
    assert c1.string_b == NON_EMPTY_1.string_b

    c2 = NON_EMPTY_2.copy()
    c2.extend_with_json_file(ONE_ONE)
    assert c2.int_a == NON_EMPTY_2.int_a
    assert c2.int_b is None
    assert c2.string_a == NON_EMPTY_2.string_a
    assert c2.string_b is None

    c3 = NON_EMPTY_3.copy()
    c3.extend_with_json_file(ONE_ONE)
    assert c3.int_a == 1
    assert c3.int_b == NON_EMPTY_3.int_b
    assert c3.string_a == "String A"
    assert c3.string_b == NON_EMPTY_3.string_b


def test_override():
    c1 = NON_EMPTY_1.copy()
    c1.extend_with_json_file(ONE_ONE, overwrite=True)
    assert c1.int_a == 1
    assert c1.int_b == NON_EMPTY_1.int_b
    assert c1.string_a == "String A"
    assert c1.string_b == NON_EMPTY_1.string_b

    c2 = NON_EMPTY_2.copy()
    c2.extend_with_json_file(ONE_ONE, overwrite=True)
    assert c2.int_a == 1
    assert c2.int_b is None
    assert c2.string_a == "String A"
    assert c2.string_b is None

    c3 = NON_EMPTY_3.copy()
    c3.extend_with_json_file(ONE_ONE, overwrite=True)
    assert c3.int_a == 1
    assert c3.int_b == NON_EMPTY_3.int_b
    assert c3.string_a == "String A"
    assert c3.string_b == NON_EMPTY_3.string_b



from steamship import Block, MimeTypes
from steamship.data.block import BlockUploadType
from steamship.invocable import InvocableResponse
from steamship.plugin.generator import Generator
from steamship.plugin.inputs.raw_block_and_tag_plugin_input import RawBlockAndTagPluginInput
from steamship.plugin.outputs.plugin_output import UsageReport
from steamship.plugin.outputs.raw_block_and_tag_plugin_output import RawBlockAndTagPluginOutput
from steamship.plugin.request import PluginRequest

TEST_BYTES_STRING = "Test MP3 bytes"


class TestGeneratorReturnsBytes(Generator):
    def run(
        self, request: PluginRequest[RawBlockAndTagPluginInput]
    ) -> InvocableResponse[RawBlockAndTagPluginOutput]:

        _bytes = TEST_BYTES_STRING.encode("utf-8")
        result = [
            Block(upload_bytes=_bytes, mime_type=MimeTypes.MP3, upload_type=BlockUploadType.FILE)
        ]

        return InvocableResponse(
            data=RawBlockAndTagPluginOutput(blocks=result, usage=[UsageReport.run_tokens(5)])
        )



import json
import random
from typing import Any, List, Union

from steamship import Block, Steamship, SteamshipError, Task
from steamship.agents.llms import OpenAI
from steamship.agents.schema import AgentContext, Tool
from steamship.agents.utils import get_llm, with_llm
from steamship.utils.repl import ToolREPL

DEFAULT_PROMPT = """INSTRUCTIONS:
Generate a JSON object describing {table_description}.
Always return a non-empty value for every field in the object.

FIELDS DESIRED:
{fields_desired}

EXAMPLE OBJECTS:
{example_objects}

NEW OBJECT:
{new_object_prefix}
"""

DEFAULT_PLURAL_OBJECT_DESCRIPTION = "employees of a company"
DEFAULT_OBJECT_KEYS = ["Name", "Age", "Gender"]
DEFAULT_OBJECT_EXAMPLES = [
    ["Bob", 30, "Male"],
    ["Susan", 32, "Female"],
    ["Zhenzhong", 40, "Male"],
    ["Luis", 32, "Male"],
    ["Roberta", 35, "Female"],
    ["Sofia", 30, "Female"],
]
DEFAULT_NEW_ROW_PREFIX_FIELDS = []


class JsonObjectGeneratorTool(Tool):
    """
    Example tool to illustrate generating a new JSON object provided a set of examples.

    This is useful as an example of how to generate a new structured object:

    - A Person (e.g. name, gender, age)
    - A Proposed Podcast Episode (e.g. title, description, tags)

    The tool takes no input at runtime: it's a true generator parameterized only at initializtion time.

    The tool's parameterization is somewhat CSV-like in construction.

    """

    rewrite_prompt: str = DEFAULT_PROMPT
    """The prompt used to generate a new JSON object."""

    plural_object_description: str = DEFAULT_PLURAL_OBJECT_DESCRIPTION
    """Plural description of the object. E.g. 'employees of a company' or 'people' or 'podcast episodes'"""

    object_keys: List[str] = DEFAULT_OBJECT_KEYS
    """The keys the JSON should have."""

    example_rows: List[List[str]] = DEFAULT_OBJECT_EXAMPLES
    """List of example object values, aligned to the `object_keys` parameter."""

    new_row_prefix_fields: List[str] = DEFAULT_NEW_ROW_PREFIX_FIELDS
    """Any fields that should be hard-coded for the new row. These must be grouped as the first set of fields."""

    shuffle_example_rows: bool = True
    """Whether randomly shuffle example rows to induce a bit of variety even with low LLM temperature."""

    validate_output_as_json: bool = True
    """Whether to validate that the output is actually JSON."""

    name: str = "JsonObjectTool"
    human_description: str = "Generates a new JSON object."
    agent_description: str = "(set at initialization time)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.agent_description = (
            f"Used to generate an instance of the {self.plural_object_description} table. "
            "Input: Anything. "
            f"Output A tab-separated row describing a new instance of the {self.plural_object_description} table."
        )

    def kv_clause(self, key: str, value: str) -> str:
        """Return an escaped, JSON style key-value clause `"key": "value"`"""
        value = str(value).replace('"', '\\"')
        clause = f'"{key}": "{value}"'
        return clause

    def object_json(self, schema: List[str], values: List[str]):
        """Render a CSV-style header row and value list into a JSON object."""
        clauses = []
        for field, value in zip(schema, values):
            clauses.append(self.kv_clause(field, value))

        return "{" + ", ".join(clauses) + "}"

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        """Ignore tool input and generate a JSON object described by the tool's configuration.

        Inputs
        ------
        input: List[Block]
            A list of blocks that will be ignored.
        memory: AgentContext
            The active AgentContext.

        Output
        ------
        output: List[Blocks]
            A single block containing a new row of the table described by the tool's configuration.
        """

        if self.shuffle_example_rows:
            # Shuffle the example rows to get a bit of variety even with low temperature.
            random.shuffle(self.example_rows)

        # Generate example JSON objects with a fixed key ordering.
        example_objects = [
            self.object_json(self.object_keys, example_row) for example_row in self.example_rows
        ]
        example_objects_str = "\n".join(example_objects)

        # Generate the new row line. At a minimum it's the `{` character, but it may also hard-code a number of
        # fields that should be affixed rather than generated.
        new_object_prefix = "{"
        for i in range(len(self.new_row_prefix_fields)):
            clause = self.kv_clause(self.object_keys[i], self.new_row_prefix_fields[i])
            new_object_prefix += f"{clause}, "

        # Compile the final generation prompt.
        prompt = self.rewrite_prompt.format(
            table_description=self.plural_object_description,
            fields_desired=", ".join(self.object_keys),
            example_objects=example_objects_str,
            new_object_prefix=new_object_prefix,
        )

        # Perform the generation
        llm = get_llm(context)
        res = llm.complete(prompt, stop="}")

        # Make sure we only generated one block; anything else violates the assumptions of this code.
        blocks_emitted = len(res)
        if blocks_emitted != 1:
            raise SteamshipError(message=f"{blocks_emitted} blocks emitted; expecting 1.")

        # The output JSON is generation prefix row, plus the generated content, plus a final } character
        # The reason we have to add the final "}" character is because we used it for the stop character
        full_json = new_object_prefix + res[0].text + "}"

        if self.validate_output_as_json:
            try:
                json.loads(full_json)
            except BaseException:
                raise SteamshipError(
                    message=f"Attempted to generate a JSON object, but did not generate valid JSON. Result: {full_json}"
                )

        res[0].text = full_json
        return res


if __name__ == "__main__":
    with Steamship.temporary_workspace() as client:
        ToolREPL(JsonObjectGeneratorTool()).run_with_client(
            client=client, context=with_llm(llm=OpenAI(client=client))
        )



from typing import Type

from steamship import Block, File, Tag
from steamship.data import DocTag, TagKind
from steamship.invocable import Config, InvocableResponse, create_handler
from steamship.invocable.plugin_service import PluginRequest
from steamship.plugin.blockifier.blockifier import Blockifier
from steamship.plugin.inputs.raw_data_plugin_input import RawDataPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput

# Note 1: this aligns with the same document in the internal Engine test.
# Note 2: This should be duplicated from the test_importer because of the way our test system will
#         bundle this into an invocable deployment package (importing won't work!)
HANDLE = "test-blockifier-plugin-v1"
TEST_H1 = "A Poem"
TEST_S1 = "Roses are red."
TEST_S2 = "Violets are blue."
TEST_S3 = "Sugar is sweet, and I love you."
TEST_DOC = f"# {TEST_H1}\n\n{TEST_S1} {TEST_S2}\n\n{TEST_S3}\n"


class DummyBlockifierPlugin(Blockifier):
    class DummyBlockifierConfig(Config):
        pass

    @classmethod
    def config_cls(cls) -> Type[Config]:
        return cls.DummyBlockifierConfig

    def run(
        self, request: PluginRequest[RawDataPluginInput]
    ) -> InvocableResponse[BlockAndTagPluginOutput]:
        return InvocableResponse(
            data=BlockAndTagPluginOutput(
                file=File(
                    blocks=[
                        Block(
                            text=TEST_H1,
                            tags=[Tag(kind=TagKind.DOCUMENT, name=DocTag.H1)],
                        ),
                        Block(
                            text=TEST_S1,
                            tags=[Tag(kind=TagKind.DOCUMENT, name=DocTag.SENTENCE)],
                        ),
                        Block(
                            text=TEST_S2,
                            tags=[Tag(kind=TagKind.DOCUMENT, name=DocTag.SENTENCE)],
                        ),
                        Block(
                            text=TEST_S3,
                            tags=[Tag(kind=TagKind.DOCUMENT, name=DocTag.PARAGRAPH)],
                        ),
                    ]
                )
            )
        )


handler = create_handler(DummyBlockifierPlugin)



from assets.plugins.generators.async_generator import OUTPUT_BLOCK_TEXT
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client


def test_e2e_generator():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "generators" / "async_generator.py"

    with deploy_plugin(client, parser_path, "generator") as (
        plugin,
        version,
        instance,
    ):
        test_doc = "Yo! Banana boy!"
        res = instance.generate(text=test_doc, append_output_to_file=False)
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 1
        assert res.output.blocks[0].text == OUTPUT_BLOCK_TEXT



import contextlib
import io
import logging
import os
import subprocess  # noqa: S404
import tempfile
import zipfile
from pathlib import Path
from typing import Any, Dict, Optional

from steamship_tests import ROOT_PATH, SRC_PATH, TEST_ASSETS_PATH

from steamship import Package, PackageInstance, PackageVersion, Steamship
from steamship.data.plugin import HostingType, Plugin
from steamship.data.plugin.plugin_instance import PluginInstance
from steamship.data.plugin.plugin_version import PluginVersion
from steamship.data.user import User


def install_dependencies(folder: str, requirements_path: Path):
    subprocess.run(  # noqa: S607,S603
        ["pip", "install", "--target", folder, "-r", str(requirements_path.resolve())],
        stdout=subprocess.PIPE,
    )
    # Write an empty requirements.txt to the test deployment.
    # Note that this is INTENTIONALLY different than the behavior of a deployable
    # deployed with the CLI, so that you can use the steamship version that you're currently editing
    with open(Path(folder) / "requirements.txt", "w") as requirements_file:
        requirements_file.write("")


def zip_deployable(file_path: Path) -> bytes:
    """Prepare and zip a Steamship plugin."""

    package_paths = [
        SRC_PATH / "steamship",
    ]

    zip_buffer = io.BytesIO()
    with zipfile.ZipFile(
        file=zip_buffer, mode="a", compression=zipfile.ZIP_DEFLATED, allowZip64=False
    ) as zip_file:
        zip_file.write(file_path, "api.py")

        # Copy in the package paths from source
        for package_path in package_paths:
            for root, _, files in os.walk(package_path):
                for file in files:
                    pypi_file = Path(root) / file
                    zip_file.write(pypi_file, pypi_file.relative_to(package_path.parent))

        # Copy in package paths from pip
        with tempfile.TemporaryDirectory() as package_dir:
            logging.info(f"Created tempdir for pip installs: {package_dir}")
            install_dependencies(
                folder=package_dir, requirements_path=ROOT_PATH / "requirements.txt"
            )
            # Now write that whole folder
            for root, _, files in os.walk(package_dir):
                for file in files:
                    pypi_file = Path(root) / file
                    zip_file.write(pypi_file, pypi_file.relative_to(package_dir))

        # Now we'll copy in the whole assets directory so that our test files can access things there..
        for root, _, files in os.walk(TEST_ASSETS_PATH):
            for file in files:
                the_file = Path(root) / file
                relative_path = the_file.relative_to(TEST_ASSETS_PATH.parent)
                zip_file.write(the_file, relative_path)

    # To generate zip files for engine tests, use the create_engine_test_assets script in Scripts

    return zip_buffer.getvalue()


@contextlib.contextmanager
def deploy_plugin(
    client: Steamship,
    py_path: Path,
    plugin_type: str,
    training_platform: Optional[HostingType] = None,
    version_config_template: Dict[str, Any] = None,
    instance_config: Dict[str, Any] = None,
    safe_load_handler: bool = False,
    wait_for_init: bool = True,
):
    plugin = Plugin.create(
        client,
        training_platform=training_platform,
        type_=plugin_type,
        transport="jsonOverHttp",
        description="A Plugin (python client steamship_tests)",
        is_public=False,
    )

    zip_bytes = zip_deployable(py_path)
    hosting_handler = "steamship.invocable.entrypoint.safe_handler" if safe_load_handler else None
    plugin_version = PluginVersion.create(
        client,
        "test-version",
        plugin_id=plugin.id,
        filebytes=zip_bytes,
        config_template=version_config_template,
        hosting_handler=hosting_handler,
    )
    # TODO: This is due to having to wait for the lambda to finish deploying.
    # TODO: We should update the task system to allow its .wait() to depend on this.
    _wait_for_version(plugin_version)

    plugin_instance = PluginInstance.create(
        client,
        plugin_id=plugin.id,
        plugin_version_id=plugin_version.id,
        config=instance_config,
    )

    assert plugin_instance.plugin_id == plugin.id
    assert plugin_instance.plugin_version_id == plugin_version.id

    _check_user(client, plugin_instance)

    if wait_for_init:
        plugin_instance.wait_for_init()

    yield plugin, plugin_version, plugin_instance

    _delete_deployable(plugin_instance, plugin_version, plugin)


@contextlib.contextmanager
def deploy_package(
    client: Steamship,
    py_path: Path,
    version_config_template: Dict[str, Any] = None,
    instance_config: Dict[str, Any] = None,
    safe_load_handler: bool = False,
    wait_for_init: bool = True,
):
    package = Package.create(client)

    zip_bytes = zip_deployable(py_path)
    hosting_handler = "steamship.invocable.entrypoint.safe_handler" if safe_load_handler else None
    version = PackageVersion.create(
        client,
        package_id=package.id,
        filebytes=zip_bytes,
        config_template=version_config_template,
        hosting_handler=hosting_handler,
    )

    _wait_for_version(version)
    package_instance = PackageInstance.create(
        client,
        package_id=package.id,
        package_version_id=version.id,
        config=instance_config,
    )

    assert package_instance.package_id == package.id
    assert package_instance.package_version_id == version.id

    _check_user(client, package_instance)

    if wait_for_init:
        package_instance.wait_for_init()

    yield package, version, package_instance

    _delete_deployable(package_instance, version, package)


def _check_user(client, instance):
    user = User.current(client)
    assert instance.user_id == user.id


def _delete_deployable(instance, version, deployable):
    instance.delete()


def _wait_for_version(version: [PackageVersion, PluginVersion]):
    assert version is not None



from .block import Block
from .embeddings import EmbeddingIndex
from .file import File
from .package import Package, PackageInstance, PackageVersion
from .plugin import Plugin, PluginInstance, PluginVersion
from .tags import DocTag, GenerationTag, Tag, TagKind, TagValueKey, TokenTag
from .workspace import Workspace

__all__ = [
    "Package",
    "PackageInstance",
    "PackageVersion",
    "Block",
    "EmbeddingIndex",
    "File",
    "GenerationTag",
    "Plugin",
    "PluginInstance",
    "PluginVersion",
    "Workspace",
    "DocTag",
    "Tag",
    "TagKind",
    "TokenTag",
    "TagValueKey",
]



from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin
from steamship_tests.utils.fixtures import get_steamship_client

from steamship.base import SteamshipError


def test_task_timeout():
    client = get_steamship_client()
    parser_path = PLUGINS_PATH / "taggers" / "plugin_parser.py"
    # TODO (enias): Use Enum for plugin type
    with deploy_plugin(client, parser_path, "tagger", safe_load_handler=True) as (
        plugin,
        version,
        instance,
    ):
        test_doc = "Hi there"
        res = instance.tag(doc=test_doc)
        try:
            res.wait(max_timeout_s=0.01, retry_delay_s=0.01)
            raise AssertionError("The call to wait() should throw")
        except SteamshipError:
            pass



from typing import List

from steamship.agents.react.output_parser import ReACTOutputParser
from steamship.agents.schema import LLM, Action, AgentContext, LLMAgent, Tool


class ReACTAgent(LLMAgent):
    """Selects actions for AgentService based on a ReACT style LLM Prompt and a configured set of Tools."""

    PROMPT = """Assistant is a large language model trained by OpenAI.
Assistant is designed to be able to assist with a wide range of tasks, from answering simple questions to providing in-depth explanations and discussions on a wide range of topics. As a language model, Assistant is able to generate human-like text based on the input it receives, allowing it to engage in natural-sounding conversations and provide responses that are coherent and relevant to the topic at hand.
Assistant is constantly learning and improving, and its capabilities are constantly evolving. It is able to process and understand large amounts of text, and can use this knowledge to provide accurate and informative responses to a wide range of questions. Additionally, Assistant is able to generate its own text based on the input it receives, allowing it to engage in discussions and provide explanations and descriptions on a wide range of topics.
Overall, Assistant is a powerful tool that can help with a wide range of tasks and provide valuable insights and information on a wide range of topics. Whether you need help with a specific question or just want to have a conversation about a particular topic, Assistant is here to assist.

TOOLS:
------
Assistant has access to the following tools:
{tool_index}
To use a tool, you MUST use the following format:
```
Thought: Do I need to use a tool? Yes
Action: the action to take, should be one of {tool_names}
Action Input: the input to the action
Observation: the result of the action
```


If you decide that you should use a Tool, you must generate the associated Action and Action Input.

Some tools will return Observations in the format of `Block(<identifier>)`. This will represent a successful completion
of that step and can be passed to subsequent tools, or returned to a user to answer their questions.

When you have a final response to say to the Human, or if you do not need to use a tool, you MUST use the format:

```
Thought: Do I need to use a tool? No
AI: [your final response here]
```

If a Tool generated an Observation that includes `Block(<identifier>)` and you wish to return it to the user, ALWAYS
end your response with the `Block(<identifier>)` observation. To do so, you MUST use the format:

```
Thought: Do I need to use a tool? No
AI: [your response with a suffix of: "Block(<identifier>)"].
```

Make sure to use all observations to come up with your final response.
You MUST include `Block(<identifier>)` segments in responses that generate images or audio.
DO NOT include `Block(<identifier>)` segments in responses that do not have generated images or audio.

Begin!

New input: {input}
{scratchpad}"""

    def __init__(self, tools: List[Tool], llm: LLM):
        super().__init__(output_parser=ReACTOutputParser(tools=tools), llm=llm, tools=tools)

    def next_action(self, context: AgentContext) -> Action:
        scratchpad = self._construct_scratchpad(context)
        tool_names = [t.name for t in self.tools]

        tool_index_parts = [f"- {t.name}: {t.agent_description}" for t in self.tools]
        tool_index = "\n".join(tool_index_parts)

        # for simplicity assume initial prompt is a single text block.
        # in reality, use some sort of formatter ?
        prompt = self.PROMPT.format(
            input=context.chat_history.last_user_message.text,
            tool_index=tool_index,
            tool_names=tool_names,
            scratchpad=scratchpad,
        )

        completions = self.llm.complete(prompt=prompt, stop="Observation:")
        return self.output_parser.parse(completions[0].text, context)

    def _construct_scratchpad(self, context):
        steps = []
        for action in context.completed_steps:
            steps.append(
                "Thought: Do I need to use a tool? Yes\n"
                f"Action: {action.tool.name}\n"
                f'Action Input: {" ".join([b.as_llm_input() for b in action.input])}\n'
                f'Observation: {" ".join([b.as_llm_input() for b in action.output])}\n'
            )
        scratchpad = "\n".join(steps)
        scratchpad += "Thought:"
        return scratchpad



import pytest
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.client import steamship_use
from steamship_tests.utils.deployables import deploy_package
from steamship_tests.utils.fixtures import get_steamship_client
from steamship_tests.utils.random import random_name

from steamship import SteamshipError


def test_use_package():
    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "demo_package.py"

    with deploy_package(client, demo_package_path) as (package, version, instance):
        # Test for infinite recursion bug
        assert package.__repr__()

        # Now let's invoke it!
        # Note: we're invoking the data at demo_package.py in the tests/assets/packages folder
        package_handle_1 = random_name()
        package_handle_2 = random_name()

        # Test that without specifying an instance handle, the user is delivered into a workspace called
        # packageHandle-default and delivered an instance called workspaceHandle-default
        #
        # These are the same semantics as workspaces: if you don't specify, you get taken to a `default` space,
        # but in this case we need to prefix with the type to avoid namespace collision within the space.
        with steamship_use(package.handle) as static_use_instance1:
            with steamship_use(package.handle, delete_workspace=False) as static_use_instance2:
                assert (
                    static_use_instance1.client.config.workspace_handle
                    == static_use_instance2.client.config.workspace_handle
                )
                assert static_use_instance1.handle == static_use_instance2.handle
                assert static_use_instance1.client.config.workspace_handle == package.handle
                assert static_use_instance1.handle == package.handle

        with steamship_use(package.handle, package_handle_1) as static_use_instance1:
            with steamship_use(package.handle, package_handle_2) as static_use_instance2:
                # Instance 1 and 2 have handles equal to their workspace handles
                assert (
                    static_use_instance1.client.config.workspace_handle
                    == static_use_instance1.handle
                )
                assert (
                    static_use_instance1.client.config.workspace_id
                    == static_use_instance1.workspace_id
                )
                assert (
                    static_use_instance2.client.config.workspace_handle
                    == static_use_instance2.handle
                )
                assert (
                    static_use_instance2.client.config.workspace_id
                    == static_use_instance2.workspace_id
                )

                # Instance 1 and 2 are in different workspaces
                assert static_use_instance1.workspace_id != static_use_instance2.workspace_id
                assert static_use_instance1.workspace_id != static_use_instance2.workspace_id

                # And neither one of these is the default workspace
                assert static_use_instance1.workspace_id != client.config.workspace_id
                assert static_use_instance2.workspace_id != client.config.workspace_id
                assert static_use_instance1.client.config.workspace_handle != "default"
                assert static_use_instance2.client.config.workspace_handle != "default"

                # And they are in the requested workspaces
                assert static_use_instance1.client.config.workspace_handle == package_handle_1
                assert static_use_instance2.client.config.workspace_handle == package_handle_2

            # We can also bring up a second instance of the same invocable
            with steamship_use(
                package.handle, package_handle_1, delete_workspace=False
            ) as static_use_instance1a:
                assert (
                    static_use_instance1a.client.config.workspace_handle
                    == static_use_instance1a.handle
                )
                assert (
                    static_use_instance1a.client.config.workspace_id
                    == static_use_instance1a.workspace_id
                )
                assert static_use_instance1a.workspace_id == static_use_instance1.workspace_id
                # And the handle is the same
                assert (
                    static_use_instance1a.handle == static_use_instance1.handle
                )  # It's the same instance (handle)
                assert (
                    static_use_instance1a.id == static_use_instance1.id
                )  # It's the same instance (id)

            # Or we could have (1) created a client anchored to the Workspace `package_handle_1` and then
            # (2) Loaded that handle from within the client.
            client2 = get_steamship_client(workspace_handle=package_handle_1)
            assert client2.config.workspace_handle == package_handle_1
            assert client2.config.workspace_id == static_use_instance1.workspace_id

            static_use_instance1a = client2.use(package.handle, package_handle_1)
            assert (
                static_use_instance1a.client.config.workspace_handle == static_use_instance1a.handle
            )  # The client is in the same workspace (handle)!
            assert (
                static_use_instance1a.client.config.workspace_id
                == static_use_instance1a.workspace_id
            )  # The client is in the same workspace (id)!
            assert (
                static_use_instance1a.workspace_id == static_use_instance1.workspace_id
            )  # It's in the same workspace!
            # And the handle is the same
            assert (
                static_use_instance1a.handle == static_use_instance1.handle
            )  # It's the same instance! (handle)
            assert (
                static_use_instance1a.id == static_use_instance1.id
            )  # It's the same instance! (id)

            # And here's the potentially hazardous thing that's possible:
            # You can use a client's member function `use` to create a second instance of that package that shares the
            # same workspace as the first, meaning it implicitly shares data.
            #
            # This is potentially useful, so it's not clear we want to forbid it (e.g. package1 could tag data, and
            # package2 could query data). But we want to encourage `Steamship.use` over `client.use` for basic use
            # due to the easier to understand scope semantics.
            package_handle_1b = random_name()
            static_use_instance1b = client2.use(package.handle, package_handle_1b)
            assert (
                static_use_instance1b.client.config.workspace_handle == static_use_instance1.handle
            )
            assert (
                static_use_instance1b.client.config.workspace_id
                == static_use_instance1.workspace_id
            )
            assert static_use_instance1b.workspace_id == static_use_instance1.workspace_id
            # But the handle isn't the same
            assert static_use_instance1b.handle != static_use_instance1.handle
            assert static_use_instance1b.id != static_use_instance1.id


def test_use_package_fails_with_same_instance_name_but_different_package_name():
    client = get_steamship_client()

    instance_handle = random_name()

    demo_package_path = PACKAGES_PATH / "demo_package.py"

    with deploy_package(client, demo_package_path) as (app, version, instance):
        with deploy_package(client, demo_package_path) as (app2, version2, instance2):
            client.use(app.handle, instance_handle)

            # Should fail because we're using the shortcut `import_plugin` method with the same instance
            with pytest.raises(SteamshipError):
                client.use(app2.handle, instance_handle)


def test_use_package_fails_with_same_instance_name_but_different_config_unless_anon():
    config_template = {
        "greeting": {"type": "string"},
        "snake_case_config": {"type": "string"},
        "camelCaseConfig": {"type": "string"},
        "defaultConfig": {"type": "string", "default": "defaulted"},
    }
    instance_config = {
        "greeting": "hi",
        "snake_case_config": "hisss",
        "camelCaseConfig": "spit!",
    }
    instance_config_2 = {
        "greeting": "hi there",
        "snake_case_config": "hisss",
        "camelCaseConfig": "spit!",
    }

    client = get_steamship_client()
    hello_world_path = PACKAGES_PATH / "configurable_hello_world.py"

    with deploy_package(
        client,
        hello_world_path,
        version_config_template=config_template,
        instance_config=instance_config,
    ) as (package, version, instance):
        # This should work
        instance_2 = client.use(package.handle, instance.handle, config=instance_config)
        assert instance_2.handle == instance.handle

        # This should fail because config changed
        with pytest.raises(SteamshipError):
            client.use(package.handle, instance.handle, config=instance_config_2)

        # This succeeds because the name includes the hash of the config
        instance_3 = client.use(package.handle, config=instance_config)
        assert instance_3.handle != package.handle

        # This succeeds because the name includes the hash of the config
        instance_4 = client.use(package.handle, config=instance_config_2)
        assert instance_4.handle != package.handle
        assert instance_4.handle != instance_3.handle



from __future__ import annotations

import pytest
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import SteamshipError
from steamship.base.model import CamelModel
from steamship.base.tasks import TaskState


class NoOpResult(CamelModel):
    pass


def test_background_task_call():
    client = get_steamship_client()

    # The No-Op API call literally does nothing.
    result_task_1 = client.post(
        "task/noop",
        expect=NoOpResult,
    )
    assert result_task_1 is not None
    assert type(result_task_1) == NoOpResult

    # When we background it, we get a task back instead
    result_2_task = client.post("task/noop", expect=NoOpResult, as_background_task=True)
    assert result_2_task is not None
    assert result_2_task.state == TaskState.waiting

    result_2_task.wait()

    # And now it has completed
    assert result_2_task.state == TaskState.succeeded
    assert result_2_task.output is not None
    assert type(result_2_task.output) == NoOpResult


def test_task_wait_milliseconds():
    client = get_steamship_client()

    # Can't wait a negative amount
    with pytest.raises(SteamshipError):
        result_task_1 = client.post("task/noop", expect=NoOpResult, task_delay_ms=-43)

    # No wait is no wait
    result_task_1 = client.post("task/noop", expect=NoOpResult, task_delay_ms=0)

    assert result_task_1 is not None
    assert type(result_task_1) == NoOpResult

    # When we background it, we get a task back instead
    result_2_task = client.post("task/noop", expect=NoOpResult, task_delay_ms=1000)
    assert result_2_task is not None
    assert result_2_task.state == TaskState.waiting

    result_2_task.wait()

    # And now it has completed
    assert result_2_task.state == TaskState.succeeded
    assert result_2_task.output is not None
    assert type(result_2_task.output) == NoOpResult


def test_task_update():
    client = get_steamship_client()

    # We'll background this operation in order to transform it into a task we can manipulate
    result_task = client.post("task/noop", expect=NoOpResult, as_background_task=True)
    assert result_task is not None
    assert result_task.state == TaskState.waiting

    result_task.wait()

    assert result_task.state == TaskState.succeeded

    # Now we've got a task that the Engine is done with. Let's test that we are permitted to manipulate it.
    # TODO(ted): Right now only a user can update their own tasks. We should limit the scope even further so that
    # a user can update their own tasks in a constrained, workflow-compatible way.

    orig_input = result_task.input
    orig_status = result_task.status_message

    input = '{"test": "input"}'
    status = "Status Message"

    result_task.input = input
    result_task.status_message = status

    # Only update the output field.
    result_task.post_update(
        fields={
            "status_message",
        }
    )

    # This will refresh the task.
    result_task.refresh()

    # The original input is unchanged; the local modifications have been overwritten by the remote state, which
    # was not updated.
    assert result_task.input == orig_input

    # The output is the new output. The remote state was updated by the client before the refresh returned
    assert result_task.status_message == status
    assert result_task.status_message != orig_status



from __future__ import annotations

import io
from enum import Enum
from typing import TYPE_CHECKING, Any, List, Optional, Type, Union

from pydantic import BaseModel, Field

from steamship import MimeTypes, SteamshipError
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import GetRequest, IdentifierRequest, ListRequest, Request, SortOrder
from steamship.base.response import ListResponse, Response
from steamship.base.tasks import Task
from steamship.data.block import Block
from steamship.data.embeddings import EmbeddingIndex
from steamship.data.tags import Tag
from steamship.utils.binary_utils import flexi_create

if TYPE_CHECKING:
    from steamship.data.operations.generator import GenerateResponse
    from steamship.data.operations.tagger import TagResponse


class FileUploadType(str, Enum):
    FILE = "file"  # A file uploaded as bytes or a string
    FILE_IMPORTER = "fileImporter"  # A fileImporter will be used to create the file
    BLOCKS = "blocks"  # Blocks are sent to create a file


class FileClearResponse(Response):
    id: str


class ListFileRequest(ListRequest):
    pass


class ListFileResponse(ListResponse):
    files: List[File]


class FileQueryRequest(Request):
    tag_filter_query: str


class File(CamelModel):
    """A file."""

    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None
    mime_type: MimeTypes = None
    workspace_id: str = None
    blocks: List[Block] = []
    tags: List[Tag] = []
    filename: str = None
    public_data: bool = False

    class CreateResponse(Response):
        data_: Any = None
        mime_type: str = None

        def __init__(
            self,
            data: Any = None,
            string: str = None,
            _bytes: Union[bytes, io.BytesIO] = None,
            json: io.BytesIO = None,
            mime_type: str = None,
        ):
            super().__init__()
            data, mime_type, encoding = flexi_create(
                data=data, string=string, json=json, _bytes=_bytes, mime_type=mime_type
            )
            self.data_ = data
            self.mime_type = mime_type

        @classmethod
        def parse_obj(cls: Type[BaseModel], obj: Any) -> Response:
            obj["data"] = obj.get("data") or obj.get("data_")
            if "data_" in obj:
                del obj["data_"]
            return super().parse_obj(obj)

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["file"] if "file" in obj else obj
        return super().parse_obj(obj)

    def delete(self) -> File:
        return self.client.post(
            "file/delete",
            IdentifierRequest(id=self.id),
            expect=File,
        )

    @staticmethod
    def get(
        client: Client,
        _id: str = None,
        handle: str = None,
    ) -> File:
        return client.post(
            "file/get",
            IdentifierRequest(id=_id, handle=handle),
            expect=File,
        )

    @staticmethod
    def create(
        client: Client,
        content: Union[str, bytes] = None,
        mime_type: MimeTypes = None,
        handle: str = None,
        blocks: List[Block] = None,
        tags: List[Tag] = None,
        public_data: bool = False,
    ) -> Any:

        if content is None and blocks is None:
            if tags is None:
                raise SteamshipError(message="Either filename, content, or tags must be provided.")
            else:
                content = ""
        if content is not None and blocks is not None:
            raise SteamshipError(
                message="Please provide only `blocks` or `content` to `File.create`."
            )

        if blocks is not None:
            upload_type = FileUploadType.BLOCKS
        elif content is not None:
            upload_type = FileUploadType.FILE
        else:
            raise Exception("Unable to determine upload type.")

        req = {
            "handle": handle,
            "type": upload_type,
            "mimeType": mime_type,
            "blocks": [
                block.dict(by_alias=True, exclude_unset=True, exclude_none=True)
                for block in blocks or []
            ],
            "tags": [
                tag.dict(by_alias=True, exclude_unset=True, exclude_none=True) for tag in tags or []
            ],
            "publicData": public_data,
        }

        file_data = (
            ("file-part", content, "multipart/form-data")
            if upload_type != FileUploadType.BLOCKS
            else None
        )

        # Defaulting this here, as opposed to in the Engine, because it is processed by Vapor
        return client.post(
            "file/create",
            payload=req,
            file=file_data,
            expect=File,
        )

    @staticmethod
    def create_with_plugin(
        client: Client,
        plugin_instance: str,
        url: str = None,
        mime_type: str = None,
    ) -> Task[File]:

        req = {
            "type": FileUploadType.FILE_IMPORTER,
            "url": url,
            "mimeType": mime_type,
            "pluginInstance": plugin_instance,
        }

        return client.post("file/create", payload=req, expect=File, as_background_task=True)

    def refresh(self) -> File:
        refreshed = File.get(self.client, self.id)
        self.__init__(**refreshed.dict())
        self.client = refreshed.client
        for block in self.blocks:
            block.client = self.client
        return self

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
    ) -> FileQueryResponse:

        req = FileQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "file/query",
            payload=req,
            expect=FileQueryResponse,
        )
        return res

    def raw(self):
        return self.client.post(
            "file/raw",
            payload=GetRequest(
                id=self.id,
            ),
            raw_response=True,
        )

    @property
    def raw_data_url(self) -> Optional[str]:
        """Return a URL at which the data content of this File can be accessed.  If public_data is True,
        this content can be accessed without an API key.
        """
        if self.client is not None:
            return f"{self.client.config.api_base}file/{self.id}/raw"
        else:
            return None

    def blockify(self, plugin_instance: str = None, wait_on_tasks: List[Task] = None) -> Task:
        from steamship.data.operations.blockifier import BlockifyRequest
        from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput

        req = BlockifyRequest(type="file", id=self.id, plugin_instance=plugin_instance)

        return self.client.post(
            "plugin/instance/blockify",
            payload=req,
            expect=BlockAndTagPluginOutput,
            wait_on_tasks=wait_on_tasks,
        )

    def tag(
        self,
        plugin_instance: str = None,
        wait_on_tasks: List[Task] = None,
    ) -> Task[TagResponse]:
        from steamship.data.operations.tagger import TagRequest, TagResponse
        from steamship.data.plugin import PluginTargetType

        req = TagRequest(type=PluginTargetType.FILE, id=self.id, plugin_instance=plugin_instance)
        return self.client.post(
            "plugin/instance/tag", payload=req, expect=TagResponse, wait_on_tasks=wait_on_tasks
        )

    def generate(
        self,
        plugin_instance_handle: str,
        start_block_index: int = None,
        end_block_index: Optional[int] = None,
        block_index_list: Optional[List[int]] = None,
        append_output_to_file: bool = True,
        options: Optional[dict] = None,
        wait_on_tasks: List[Task] = None,
        make_output_public: bool = False,
    ) -> Task[GenerateResponse]:
        """Generate new content from this file. Assumes this file as context for input and output.  May specify start and end blocks."""
        from steamship.data.operations.generator import GenerateRequest, GenerateResponse

        if append_output_to_file:
            output_file_id = self.id
        else:
            output_file_id = None

        req = GenerateRequest(
            plugin_instance=plugin_instance_handle,
            input_file_id=self.id,
            input_file_start_block_index=start_block_index,
            input_file_end_block_index=end_block_index,
            input_file_block_index_list=block_index_list,
            append_output_to_file=append_output_to_file,
            output_file_id=output_file_id,
            options=options,
            make_output_public=make_output_public,
        )
        return self.client.post(
            "plugin/instance/generate", req, expect=GenerateResponse, wait_on_tasks=wait_on_tasks
        )

    def index(self, plugin_instance: Any = None) -> EmbeddingIndex:
        """Index every block in the file.

        TODO(ted): Enable indexing the results of a tag query.
        TODO(ted): It's hard to load the EmbeddingIndexPluginInstance with just a handle because of the chain
        of things that need to be created to it to function."""

        # Preserve the prior behavior of embedding the full text of each block.
        tags = [
            Tag(text=block.text, file_id=self.id, block_id=block.id, kind="block")
            for block in self.blocks or []
        ]
        return plugin_instance.insert(tags)

    @staticmethod
    def list(
        client: Client,
        page_size: Optional[int] = None,
        page_token: Optional[str] = None,
        sort_order: Optional[SortOrder] = SortOrder.DESC,
    ) -> ListFileResponse:
        return client.post(
            "file/list",
            ListFileRequest(pageSize=page_size, pageToken=page_token, sortOrder=sort_order),
            expect=ListFileResponse,
        )

    def append_block(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
        public_data: bool = False,
    ) -> Block:
        """Append a new block to this File.  This is a convenience wrapper around
        Block.create(). You should provide only one of text, content, or url.

        This is a server-side operation, saving the new Block to the file. The new block
        is appended to this client-side File as well for convenience.
        """
        block = Block.create(
            self.client,
            file_id=self.id,
            text=text,
            tags=tags,
            content=content,
            url=url,
            mime_type=mime_type,
            public_data=public_data,
        )
        self.blocks.append(block)
        return block

    def set_public_data(self, public_data: bool):
        """Set the public_data flag on this File. If this object already exists server-side, update the flag."""
        self.public_data = public_data
        if self.client is not None and self.id is not None:
            req = {
                "id": self.id,
                "publicData": self.public_data,
            }
            return self.client.post("file/update", payload=req, expect=File)


class FileQueryResponse(Response):
    files: List[File]


ListFileResponse.update_forward_refs()



from __future__ import annotations

from steamship.plugin.outputs.plugin_output import PluginOutput


class TrainPluginOutput(PluginOutput):
    """
    This is the object produced by a completed trainable operation, stored as the `output` field of a `train` task.
    """

    # The PluginInstance ID being trained
    plugin_instance_id: str = None

    # This should always represent the most recent snapshot of the model in Steamship
    # It is the output of ModelCheckpoint.archive_path_in_steamship
    archive_path: str = None

    # Arbitrary key-valued data to provide to the `run` method when this plugin is Run.
    inference_params: dict = None

    # Arbitrary key-valued data to provide information about training status or training results.
    training_progress: dict = None  # For tracking the progress (e.g. 3 / 40 epochs completed)
    training_results: dict = None  # For tracking accuracy (e.g. f1=0.8)



import os
import zipfile
from abc import ABC, abstractmethod
from pathlib import Path

import click
from semver import VersionInfo

from steamship import Package, PackageVersion, PluginVersion, Steamship, SteamshipError
from steamship.cli.manifest_init_wizard import validate_handle, validate_version_handle
from steamship.cli.ship_spinner import ship_spinner
from steamship.cli.utils import find_api_py, get_api_module
from steamship.data import Plugin
from steamship.data.manifest import Manifest
from steamship.data.user import User
from steamship.invocable.lambda_handler import get_class_from_module

DEFAULT_BUILD_IGNORE = [
    "build",
    ".git",
    ".venv",
    ".ipynb_checkpoints",
    ".DS_Store",
    "venv",
    "tests",
    "examples",
    ".idea",
    "__pycache__",
]


def update_config_template(manifest: Manifest):
    path = find_api_py()
    api_module = get_api_module(path)
    invocable_type = get_class_from_module(api_module)

    config_parameters = invocable_type.config_cls().get_config_parameters()

    if manifest.configTemplate != config_parameters:
        if len(config_parameters) > 0:
            click.secho("Config parameters changed; updating steamship.json.", fg="cyan")
            for param_name, param in config_parameters.items():
                click.echo(f"{param_name}:")
                click.echo(f"\tType: {param.type}")
                click.echo(f"\tDefault: {param.default}")
                click.echo(f"\tDescription: {param.description}")
        else:
            click.secho("Config parameters removed; updating steamship.json.", fg="cyan")

    manifest.configTemplate = config_parameters
    manifest.save()


def get_archive_path(manifest: Manifest) -> Path:
    return Path(".") / "build" / "archives" / f"{manifest.handle}_v{manifest.version}.zip"


def bundle_deployable(manifest: Manifest):
    archive_path = get_archive_path(manifest)
    archive_path.parent.mkdir(parents=True, exist_ok=True)
    excludes = DEFAULT_BUILD_IGNORE + manifest.build_config.get("ignore", [])

    archive_path.unlink(missing_ok=True)

    # This zipfile packaging is modeled after the typescript CLI.
    # Items in non-excluded root folders are moved to the top-level.

    with zipfile.ZipFile(
        file=archive_path, mode="a", compression=zipfile.ZIP_DEFLATED, allowZip64=False
    ) as zip_file:

        root = Path(".")
        for file_path in root.iterdir():
            if file_path.name not in excludes:
                if file_path.is_dir():
                    for directory, _, files in os.walk(file_path):
                        subdirectory_path = Path(directory)
                        if Path(directory).name not in excludes:
                            for file in files:
                                pypi_file = subdirectory_path / file
                                relative_to = pypi_file.relative_to(file_path)
                                zip_file.write(pypi_file, relative_to)

                else:
                    zip_file.write(file_path)


class DeployableDeployer(ABC):
    @abstractmethod
    def _create_version(self, client: Steamship, manifest: Manifest, thing_id: str):
        pass

    @abstractmethod
    def create_object(self, client: Steamship, manifest: Manifest):
        pass

    @abstractmethod
    def update_object(self, deployable, client: Steamship, manifest: Manifest):
        pass

    @abstractmethod
    def deployable_type(self):
        pass

    def create_or_fetch_deployable(self, client: Steamship, user: User, manifest: Manifest):
        if not manifest.handle or len(manifest.handle) == 0:
            self.ask_for_new_handle(manifest, was_missing=True)

        deployable = None
        while deployable is None:
            click.echo(
                f"Creating / fetching {self.deployable_type()} with handle [{manifest.handle}]... ",
                nl=False,
            )
            try:
                deployable = self.create_object(client, manifest)
                if deployable.user_id != user.id:
                    self.ask_for_new_handle(manifest)
                    deployable = None
            except SteamshipError as e:
                if e.message == "Something went wrong.":
                    self.ask_for_new_handle(manifest)
                else:
                    click.secho(
                        f"Unable to create / fetch {self.deployable_type()}. Server returned message: {e.message}"
                    )
                    click.get_current_context().abort()

        self.update_object(deployable, client, manifest)

        click.echo("Done.")
        return deployable

    def ask_for_new_handle(self, manifest: Manifest, was_missing: bool = False):
        if not was_missing:
            try_again = click.confirm(
                click.style(
                    f"\nIt looks like that handle [{manifest.handle}] is already in use. Would you like to change the handle and try again?",
                    fg="yellow",
                ),
                default=True,
            )
            if not try_again:
                click.get_current_context().abort()

        new_handle = click.prompt(
            f"What handle would you like to use for your {self.deployable_type()}? Valid characters are a-z and -",
            value_proc=validate_handle,
        )
        manifest.handle = new_handle
        manifest.save()

    def create_version(self, client: Steamship, manifest: Manifest, thing_id: str):
        version = None

        if not manifest.version or len(manifest.version) == 0:
            self.ask_for_new_version_handle(manifest, was_missing=True)

        while version is None:
            click.echo(f"Deploying version {manifest.version} of [{manifest.handle}]... ", nl=False)
            try:
                with ship_spinner():
                    version = self._create_version(client, manifest, thing_id)
            except SteamshipError as e:
                if "The object you are trying to create already exists." in e.message:
                    self.ask_for_new_version_handle(manifest)
                else:
                    click.secho(f"\nUnable to deploy {self.deployable_type()} version.", fg="red")
                    click.secho(f"Server returned message: {e.message}")
                    if "ModuleNotFoundError" in e.message:
                        click.secho(
                            "It looks like you may be missing a dependency in your requirements.txt.",
                            fg="yellow",
                        )
                    click.get_current_context().abort()
        click.echo("\nDone. 🚢")

    def ask_for_new_version_handle(self, manifest: Manifest, was_missing: bool = False):
        if not was_missing:
            try_again = click.confirm(
                click.style(
                    f"\nIt looks like that version [{manifest.version}] has already been deployed. Would you like to change the version handle and try again?",
                    fg="yellow",
                ),
                default=True,
            )
            if not try_again:
                click.get_current_context().abort()

        default_new = "1.0.0"
        try:
            default_new = str(VersionInfo.parse(manifest.version).bump_prerelease())
        except ValueError:
            pass
        old_archive_path = get_archive_path(manifest)
        new_version_handle = click.prompt(
            "What should the new version be? Valid characters are a-z, 0-9, . and -",
            value_proc=validate_version_handle,
            default=default_new,
        )
        manifest.version = new_version_handle
        manifest.save()
        new_archive_path = get_archive_path(manifest)
        old_archive_path.rename(new_archive_path)


class PackageDeployer(DeployableDeployer):
    def _create_version(self, client: Steamship, manifest: Manifest, thing_id: str):
        return PackageVersion.create(
            client=client,
            config_template=manifest.config_template_as_dict(),
            handle=manifest.version,
            filename=f"build/archives/{manifest.handle}_v{manifest.version}.zip",
            package_id=thing_id,
        )

    def create_object(self, client: Steamship, manifest: Manifest):
        return Package.create(
            client,
            handle=manifest.handle,
            profile=manifest,
            is_public=manifest.public,
            fetch_if_exists=True,
        )

    def update_object(self, deployable, client: Steamship, manifest: Manifest):
        deployable.profile = manifest

        package = deployable.update(client)
        return package

    def deployable_type(self):
        return "package"


class PluginDeployer(DeployableDeployer):
    def _create_version(self, client: Steamship, manifest: Manifest, thing_id: str):
        return PluginVersion.create(
            client=client,
            config_template=manifest.config_template_as_dict(),
            handle=manifest.version,
            filename=f"build/archives/{manifest.handle}_v{manifest.version}.zip",
            plugin_id=thing_id,
        )

    def create_object(self, client: Steamship, manifest: Manifest):
        return Plugin.create(
            client,
            description=manifest.description,
            is_public=manifest.public,
            transport=manifest.plugin.transport,
            type_=manifest.plugin.type,
            handle=manifest.handle,
            fetch_if_exists=True,
        )

    def update_object(self, deployable, client: Steamship, manifest: Manifest):
        deployable.profile = manifest

        plugin = deployable.update(client)
        return plugin

    def deployable_type(self):
        return "plugin"
